#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <linux/types.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <net/if.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <arpa/inet.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <math.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/time.h>
#include <time.h>

/*#include <zlib.h>*/

#include "x10-i2c.h"

#define X10I2C_MAIN_BUS   2
#define X10I2C_BBONE_EEPROM_BUS 3
#define X10I2C_SITARA_BUS 1

#define PCA9544ADDR(lport_num) (0x71+((lport_num-1)>>2))
#define PCA9544VALUE_SFP(lport_num) ((lport_num-1) - (((lport_num-1)>>2)<<2) + 4 )
#define PCA9544VALUE_EXP(lport_num) (PCA9544VALUE_SFP(lport_num)&0xFE)

#define PCA9544ADDR_RS(lport_num) (0x72 + (((lport_num-1)>>3)<<1))
#define PCA9544VALUE_EXP_RS 0x07
#define PCA9544VALUE_PCA9552LEDS 0x06
#define TCA9535ADDR_RS 0x27

#define TCA9535ADDR 0x20
#define SFPADDR0 0x50
#define SFPADDR1 0x51
#define ADS7830ADDR 0x49
#define SITARA_ADS7830ADDR1 0x48
#define SITARA_ADS7830ADDR2 0x49

#define SITARA_I2C_MULTIPLEXOR 0x70
#define SITARA_I2C_FUJITSU     0x51
#define SITARA_I2C_EEPROM      0x51

#define SITARA_SYSIO_TCA9535ADDR   0x27
#define SITARA_MAIN_ADT7463      0x2D
#define SITARA_CHASSIS_ADT7463   0x2E
#define SITARA_LCD_ADDR 0x3C

#define SITARA_SYSIO_TCA9535_BACKLIGHT    0x1
#define SITARA_SYSIO_TCA9535_PSB0_PRESENT 0x2
#define SITARA_SYSIO_TCA9535_PSB1_PRESENT 0x4
#define SITARA_SYSIO_TCA9535_PWR0_GOOD    0x10
#define SITARA_SYSIO_TCA9535_PWR0_FAULT   0x20
#define SITARA_SYSIO_TCA9535_PWR1_GOOD    0x40
#define SITARA_SYSIO_TCA9535_PWR1_FAULT   0x80

#define PCA9552_ADDR 0x60
#define PCA9552_REG_PSC0 2
#define PCA9552_REG_PWM0 3
#define PCA9552_REG_PSC1 4
#define PCA9552_REG_PWM1 5
#define PCA9552_REG_LS0 6

#define GPIO_RESET    48
#define GPIO_RESET_SITARA    82
#define GPIO_POWERLED 39
#define GPIO_POWERLED_SITARA2 106
#define GPIO_EPROMWP_SITARA2 3

#define X10I2C_DEBUG_BASE_LEVEL 1

#ifdef X10I2C_DEBUG
#define X10I2CDBGL(level, format, ...) debug_print(&(h->dbg), X10I2C_DEBUG_BASE_LEVEL+level, " %s: "format, __FUNCTION__, ##__VA_ARGS__)
#define X10I2CDBG(format, ...) X10I2CDBGL(X10I2C_DEBUG_BASE_LEVEL, format, ##__VA_ARGS__)
#define X10I2CERR(format, ...) X10I2CDBGL(0, "ERR: "format, ##__VA_ARGS__)
#else
#define X10I2CERR(format, ...)
#define X10I2CDBG(format, ...)
#define X10I2CDBGL(level, format, ...)
#endif

int x10_init_gpio(x10i2c_t *h);
int x10_init_pwm(x10i2c_t *h);
int x10i2c_set_pca9552leds(x10i2c_t *h, uint8_t lp, uint8_t leds);
int x10i2c_init_ADT7463(x10i2c_t *h, uint8_t *table, uint16_t tsize, uint8_t i2caddr);
int x10i2c_get_sitara2_temps(x10i2c_t *h, uint16_t *temps);
int x10i2c_get_sitara2_fans(x10i2c_t *h, uint16_t *fans);
int x10i2c_display_init(x10i2c_t *h, uint8_t i2caddr);
int x10i2c_init_pca9552(x10i2c_t *h);

uint8_t init_ADT7463main_table[]={
  0x40, 0x01,     /* Startup monitoring */
  0x73, 0x0E,     /* TCH2 - TCH3 as AIN */
  0x7b, 0xFF,     /* 4 pulses per revolution for all fans */
  0x7d, 0x04,     /* 130 mV treshold */
  0x5C, 0xE2,     /* Manual mode */
  0x5D, 0xE2,     /* */
  0x5E, 0xE2,     /* */
  0x31, 0x90,
  0x32, 0x90,
  0x62, 0x10,     /* aucoustic sync */
  0x78, 0xF2,     /* some magick with nTHERM */
  0x64, 0x40,     /* set PWM1min to 25% */
  0x65, 0x40,     /* set PWM2min to 25% */
  0x66, 0x40,     /* set PWM3min to 25% */
  0x68, 0x23,     /* set Tmin to 35C */
  0x60, 0xa4      /* set Trange to 20C */
};

uint8_t init_ADT7463chassis_table[]={
  0x40, 0x01,     /* Startup monitoring */
  0x73, 0x0F,     /* TCH1 - TCH4 as AIN */
  0x7b, 0x55,     /* 2 pulses per revolution for all fans */
  0x7d, 0x04,     /* 130 mV treshold */
  0x5C, 0xE2,     /* Manual mode */
  0x5D, 0xE2,     /* */
  0x5E, 0xE2,     /* */
  0x30, 0x30,
  0x31, 0x30,
  0x62, 0x10,     /* aucoustic sync */
  0x78, 0xF0,     /* some magick with nTHERM */
  0x64, 0x40,     /* set PWM1min to 25% */
  0x65, 0x40,     /* set PWM2min to 25% */
  0x66, 0x40,     /* set PWM3min to 25% */
  0x68, 0x23,     /* set Tmin to 35C */
  0x60, 0xa4      /* set Trange to 20C */
};


int x10_detect_i2c_bus_name(int i2cbus, char *filename, int fnsize)
{
  struct stat s;
  snprintf(filename, fnsize, "/dev/i2c/%d", i2cbus);

  if (stat(filename, &s))
    snprintf(filename, fnsize, "/dev/i2c-%d", i2cbus);

  if (stat(filename, &s))
    return -1;

  return 0;
}

int x10i2c_init(x10i2c_t **h, uint8_t model, int l)
{
  char *modelstr;
/*  uint8_t i2cbus;*/
  (*h) = malloc(sizeof(x10i2c_t));
  memset((*h), 0, sizeof(x10i2c_t));

  (*h)->fd = -1;  
  (*h)->x10model =  1;

/*  i2cbus = X10I2C_MAIN_BUS;*/

  if (!model) {
    modelstr = getenv("X10MODEL");
    if (modelstr)
      (*h)->x10model = (uint8_t)strtoul(modelstr, 0, 10);
  }
/*
  if (x10_detect_i2c_bus_name(i2cbus, (*h)->filename, sizeof((*h)->filename))) {
    free(*h);
    return -1;
  }
*/
  if (debug_init(&((*h)->dbg), l)) {
    free(*h);
    return -1;
  }

  #ifdef X10I2C_DEBUG
  debug_print(&((*h)->dbg), X10I2C_DEBUG_BASE_LEVEL, "%s model: %u\n", __FUNCTION__, (*h)->x10model);
  #endif

  return 0;
}

int x10i2c_free(x10i2c_t *h)
{
  debug_free(&h->dbg);
  free(h);
  return 0;
}

/* from i2ctools */

static inline __s32 i2c_smbus_access(int file, char read_write, __u8 command,
                                     int size, union i2c_smbus_data *data)
{
	struct i2c_smbus_ioctl_data args;

	args.read_write = read_write;
	args.command = command;
	args.size = size;
	args.data = data;
	return ioctl(file,I2C_SMBUS,&args);
}

static inline __s32 i2c_smbus_read_byte_data(int file, __u8 command)
{
	union i2c_smbus_data data;
	if (i2c_smbus_access(file,I2C_SMBUS_READ,command,
	                     I2C_SMBUS_BYTE_DATA,&data))
		return -1;
	else
		return 0x0FF & data.byte;
}

static inline __s32 i2c_smbus_write_byte_data(int file, __u8 command,
                                              __u8 value)
{
	union i2c_smbus_data data;
	data.byte = value;
	return i2c_smbus_access(file,I2C_SMBUS_WRITE,command,
	                        I2C_SMBUS_BYTE_DATA, &data);
}

int openi2c(x10i2c_t *h, int i2cbus)
{
  	unsigned long funcs;

    X10I2CDBG("\n");

    if (x10_detect_i2c_bus_name(i2cbus, h->filename, sizeof(h->filename)))
      return -1;

	h->fd = open(h->filename, O_RDWR);

	if (h->fd < 0) {
      X10I2CERR("Could not open file "
   	             "`%s': %s\n", h->filename, strerror(errno));
      return -1;
    }

	/* check adapter functionality */
	if (ioctl(h->fd, I2C_FUNCS, &funcs) < 0) {
        X10I2CERR("Could not get the adapter "
			"functionality matrix: %s\n", strerror(errno));
		close(h->fd);
		return -1;
	}

	if ( (!(funcs & I2C_FUNC_SMBUS_WRITE_BYTE)) &&
	     (!(funcs & I2C_FUNC_SMBUS_WRITE_BYTE_DATA)) ) {
	  X10I2CERR("I2C functionality test failed\n");
	  close(h->fd);
	  return -1;
	}

    /*if (flock(h->fd, LOCK_EX))*/
    if (lockf(h->fd, F_LOCK,0))
      X10I2CDBG("Can't lock '%s'\n", h->filename);

    h->current_bus = i2cbus;

	return 0;
}


int closei2c(x10i2c_t *h)
{
  X10I2CDBG("\n");
  if (h->fd>=0) {
    /*if (flock(h->fd, LOCK_UN))*/
    if (lockf(h->fd, F_ULOCK,0))
      X10I2CERR("Can't unlock '%s'\n", h->filename);
    close(h->fd);
    h->fd = -1;
  }
  return 0;
}

int set_slave_addr(x10i2c_t *h, uint8_t i2caddr)
{
  if (ioctl(h->fd, I2C_SLAVE_FORCE, i2caddr) < 0) {
    X10I2CERR("Can't set i2c address to 0x%02x: %s\n",
                i2caddr, strerror(errno));
    return -1;
  }
  return 0;
}

int get_pca9544_multiplexer(x10i2c_t *h, uint8_t i2caddr, uint8_t *val)
{
  X10I2CDBG("dev: %02X\n",
             (uint16_t)i2caddr);

  /* set slave address */
  if (set_slave_addr(h, i2caddr))
   return -1;

  /* read data */
  if (read(h->fd, val, 1)!=1) {
	X10I2CERR("Can not read from i2c device 0x%02x: %s\n",
                (uint16_t)i2caddr, strerror(errno));
  }

  return 0;
}

int set_pca9544_multiplexer(x10i2c_t *h, uint8_t i2caddr, uint8_t out)
{
  X10I2CDBG("dev: %02X out: %02X \n",
             (uint16_t)i2caddr, (uint16_t)out);

  /* set slave address */
  if (set_slave_addr(h, i2caddr))
   return -1;

  /* write data */
  if (i2c_smbus_write_byte_data(h->fd, out, out) < 0) {
	X10I2CERR("Can not write to i2c device 0x%02x: %s\n",
                (uint16_t)i2caddr, strerror(errno));
    return -1;
  }

  return 0;
}

int switch_sitara_i2c_bus(x10i2c_t *h, uint8_t i2cbus)
{
  return set_pca9544_multiplexer(h, SITARA_I2C_MULTIPLEXOR, (i2cbus)?5:4);
}

int change_sitara_i2c_bus(x10i2c_t *h, uint8_t i2cbus)
{
  if (get_pca9544_multiplexer(h, SITARA_I2C_MULTIPLEXOR, &h->sitara_bus))
    return -1;
  return set_pca9544_multiplexer(h, SITARA_I2C_MULTIPLEXOR, (i2cbus)?5:4);
}

int restore_sitara_i2c_bus(x10i2c_t *h)
{
  return set_pca9544_multiplexer(h, SITARA_I2C_MULTIPLEXOR, h->sitara_bus);
}

int set_tca9535_expander(x10i2c_t *h, uint8_t i2caddr,
                         uint8_t reg, uint8_t data, uint8_t mask)
{
  int old;

  /* X10I2CDBG("dev: %02X reg: %02X data: %02X mask %02X\n",
             (uint16_t)i2caddr, (uint16_t)reg,
              (uint16_t)data, (uint16_t)mask); */

  /* set slave address */
  if (set_slave_addr(h, i2caddr))
   return -1;

  /* read old data */
  old = i2c_smbus_read_byte_data(h->fd, reg);
  if (old < 0) {
    X10I2CERR("%s: Can not read old data\n",__FUNCTION__);
    return -1;
  }

  X10I2CDBG("dev: %02X reg: %02X old: %02X new: %02X\n",
             (uint16_t)i2caddr, (uint16_t)reg,
              (uint16_t)old, (uint16_t)(((uint8_t)old)&(~mask))|data);

  /* write data */
  if (i2c_smbus_write_byte_data(h->fd, reg, (((uint8_t)old)&(~mask))|(data & mask))<0) {
	X10I2CERR("Can not write to i2c device 0x%02x: %s\n",
                (uint16_t)i2caddr, strerror(errno));
    return -1;
  }
  return 0;
}

int get_tca9535_expander(x10i2c_t *h, uint8_t i2caddr,
                         uint8_t reg, uint8_t *data)
{
  int r=0;

  /* set slave address */
  if (set_slave_addr(h, i2caddr))
   return -1;

  /* read old data */
  r = i2c_smbus_read_byte_data(h->fd, reg);
  if (r < 0) {
    X10I2CERR("%s: Can't read old data\n",__FUNCTION__);
    return -1;
  }

  X10I2CDBG("dev: %02X reg: %02X data: %02X\n",
             (uint16_t)i2caddr, (uint16_t)reg, r);

  (*data) = (uint8_t)r;

  return 0;
}

int x10i2c_init_port_tca9535_expander(x10i2c_t *h, uint8_t i2caddr)
{
  if (h->x10model<3) {
    /* setup config registers
    * pin (0,4,5,6) and (8, 17, 18, 19) - is output
    * other inputs */
    set_tca9535_expander(h, i2caddr, 6, 0x8E, 0xFF);
    set_tca9535_expander(h, i2caddr, 7, 0x8E, 0xFF);
  } else {
    /* setup config registers
    * pin (0,4,5) and (8, 17, 18) - is output
    * other inputs */
    set_tca9535_expander(h, i2caddr, 6, 0xCE, 0xFF);
    set_tca9535_expander(h, i2caddr, 7, 0xCE, 0xFF );
  }
  /* setup polarity reverse registers */
  set_tca9535_expander(h, i2caddr, 4, 0xFF, 0xFF);
  set_tca9535_expander(h, i2caddr, 5, 0xFF, 0xFF);
  /* setup output port registers def*/
  set_tca9535_expander(h, i2caddr, 2, 0xFF, 0xFF);
  set_tca9535_expander(h, i2caddr, 3, 0xFF, 0xFF);

  return 0;
}

int x10i2c_init_sysio_tca9535_expander(x10i2c_t *h)
{
  set_tca9535_expander(h, SITARA_SYSIO_TCA9535ADDR, 6, 0xFE, 0xFF);
  set_tca9535_expander(h, SITARA_SYSIO_TCA9535ADDR, 7, 0xFE, 0xFF );
  /* setup polarity reverse registers */
  set_tca9535_expander(h, SITARA_SYSIO_TCA9535ADDR, 4, 0xFF, 0xFF);
  set_tca9535_expander(h, SITARA_SYSIO_TCA9535ADDR, 5, 0xFF, 0xFF);
  /* setup output port registers def*/
  set_tca9535_expander(h, SITARA_SYSIO_TCA9535ADDR, 2, 0xFF, 0xFF);
  set_tca9535_expander(h, SITARA_SYSIO_TCA9535ADDR, 3, 0xFF, 0xFF);
  return 0;
}

int x10i2c_init_rs_tca9535_expander(x10i2c_t *h)
{
  /* setup config registers
   * all pins as outputs */
  set_tca9535_expander(h, TCA9535ADDR_RS, 6, 0x0, 0xFF);
  set_tca9535_expander(h, TCA9535ADDR_RS, 7, 0x0, 0xFF );
  /* setup polarity reverse registers */
  set_tca9535_expander(h, TCA9535ADDR_RS, 4, 0xFF, 0xFF);
  set_tca9535_expander(h, TCA9535ADDR_RS, 5, 0xFF, 0xFF);
  /* setup output port registers def*/
  set_tca9535_expander(h, TCA9535ADDR_RS, 2, 0xFF, 0xFF);
  set_tca9535_expander(h, TCA9535ADDR_RS, 3, 0xFF, 0xFF);

  return 0;
}


int x10i2c_init_hw(x10i2c_t *h)
{
  int r;
  uint8_t i;

  if (x10_init_gpio(h))
    goto error;

  if ( (h->x10model==X10MODEL_BBONE) && (x10_init_pwm(h)) )
    goto error;


  if (openi2c(h, X10I2C_MAIN_BUS))
    goto error;  

  chmod(h->filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);

  for (i=0x71; i<0x76; ++i)
    if (set_pca9544_multiplexer(h, i, 0))
      goto error;

  /* initialize TCA9535 expanders */

  for (i = 1; i <= 24; i+=2) {

   /* setup multiplexer to attach nessessary line */
   if (set_pca9544_multiplexer(h, PCA9544ADDR(i), PCA9544VALUE_EXP(i)))
     goto error;

   /* initialize expander */
   r = x10i2c_init_port_tca9535_expander(h, TCA9535ADDR);

   /* detach i2c line back */
   set_pca9544_multiplexer(h, PCA9544ADDR(i), 0);

   if (r)
     goto error;

  }


  if (h->x10model<=X10MODEL_SITARA)

  for ( i = 1; i <= 24; i+=8 ) {

   /* setup multiplexer to attach nessessary line */
   if (set_pca9544_multiplexer(h, PCA9544ADDR_RS(i), PCA9544VALUE_EXP_RS))
     goto error;

   /* initialize expander */
   r = x10i2c_init_rs_tca9535_expander(h);

   /* detach i2c line back */
   set_pca9544_multiplexer(h, PCA9544ADDR_RS(i), 0);

   if (r)
     goto error;

  }


  if (h->x10model>=X10MODEL_SITARA2) {

    for ( i = 1; i <= 24; i+=4 ) {

     /* setup multiplexer to attach nessessary line */
     if (set_pca9544_multiplexer(h, PCA9544ADDR(i), PCA9544VALUE_PCA9552LEDS))
       goto error;

     /* initialize LEDs chip */
     r = x10i2c_init_pca9552(h);

     /* detach i2c line back */
     set_pca9544_multiplexer(h, PCA9544ADDR(i), 0);

     if (r)
       goto error;

    }

  }

  closei2c(h);


  /* Init devises on service bus */
  if (h->x10model>=X10MODEL_SITARA2) {
    if (openi2c(h, X10I2C_SITARA_BUS))
      goto error;

    chmod(h->filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);

    if (switch_sitara_i2c_bus(h, 1))
      goto error;

    /* Init sysio expander */
    if (x10i2c_init_sysio_tca9535_expander(h))
      goto error;

    /* Init fan & temp controllers */
    if (x10i2c_init_ADT7463(h, init_ADT7463main_table,
                            sizeof(init_ADT7463main_table),
                            SITARA_MAIN_ADT7463))
      goto error;

    if (x10i2c_init_ADT7463(h, init_ADT7463chassis_table,
                            sizeof(init_ADT7463chassis_table),
                            SITARA_CHASSIS_ADT7463))
      goto error;

    /* Init LCD display */
    x10i2c_display_init(h, SITARA_LCD_ADDR);

    closei2c(h);
  }

  /* switch to default i2c bus on sitara */
  if (h->x10model>=X10MODEL_SITARA) {
    if (openi2c(h, X10I2C_SITARA_BUS))
      goto error;
    if (switch_sitara_i2c_bus(h, 0))
      goto error;
    closei2c(h);
  }

  return 0;

error:
  closei2c(h);

  return -1;
}

/* manage pins on ports expanders */

int x10i2c_set_port_outs(x10i2c_t *h, uint8_t lp, uint8_t data, uint8_t mask)
{
  X10I2CDBG("prt: %02d data: %02d mask: %02X \n",
             (uint16_t)lp, (uint16_t)data, (uint16_t)mask);

  if (openi2c(h, X10I2C_MAIN_BUS))
    goto error;

  /* setup multiplexer to attach nessessary line */
  if (set_pca9544_multiplexer(h, PCA9544ADDR(lp), PCA9544VALUE_EXP(lp)))
    goto error;

  /* write to output registers 2 or 3 according to port number */
  if (set_tca9535_expander(h, TCA9535ADDR, (lp&1)?2:3, data, mask))
    goto error;

  /* detach i2c line back */
  set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);

  closei2c(h);

  return 0;

error:
  if (h->fd >= 0) {
    set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);
    closei2c(h);
  }

  return -1;
}


int x10i2c_get_port_inputs(x10i2c_t *h, uint8_t lp, uint8_t mask, uint8_t *data)
{
  X10I2CDBG("prt: %02d mask: %02X \n",
             (uint16_t)lp, (uint16_t)mask);

  if (openi2c(h, X10I2C_MAIN_BUS))
    goto error;

  /* setup multiplexer to attach nessessary line */
  if (set_pca9544_multiplexer(h, PCA9544ADDR(lp), PCA9544VALUE_EXP(lp)))
    goto error;

  /* write to output registers 2 or 3 according to port number */
  if (get_tca9535_expander(h, TCA9535ADDR, (lp&1)?0:1, data))
    goto error;

  (*data) &= mask;

  X10I2CDBG("prt: %02d mask: %02X result: %02X \n",
             (uint16_t)lp, (uint16_t)mask, (uint16_t)(*data));

  /* detach i2c line back */
  set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);

  closei2c(h);

  return 0;

error:
  if (h->fd >= 0) {
    set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);
    closei2c(h);
  }

  return -1;
}

int x10i2c_get_port_outs(x10i2c_t *h, uint8_t lp, uint8_t mask, uint8_t *data)
{
  X10I2CDBG("prt: %02d mask: %02X \n",
             (uint16_t)lp, (uint16_t)mask);

  if (openi2c(h, X10I2C_MAIN_BUS))
    goto error;

  /* setup multiplexer to attach nessessary line */
  if (set_pca9544_multiplexer(h, PCA9544ADDR(lp), PCA9544VALUE_EXP(lp)))
    goto error;

  /* write to output registers 2 or 3 according to port number */
  if (get_tca9535_expander(h, TCA9535ADDR, (lp&1)?2:3, data))
    goto error;

  (*data) &= mask;

  X10I2CDBG("prt: %02d mask: %02X result: %02X \n",
             (uint16_t)lp, (uint16_t)mask, (uint16_t)(*data));

  /* detach i2c line back */
  set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);

  closei2c(h);

  return 0;

error:
  if (h->fd >= 0) {
    set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);
    closei2c(h);
  }

  return -1;
}



int x10i2c_set_port_rs_outs(x10i2c_t *h, uint8_t lp, uint8_t data, uint8_t mask)
{
  if (openi2c(h, X10I2C_MAIN_BUS))
    goto error;

  /* setup multiplexer to attach nessessary line */
  if (set_pca9544_multiplexer(h, PCA9544ADDR_RS(lp), PCA9544VALUE_EXP_RS))
    goto error;

  /* write to output registers 2 or 3 according to port number */
  if (set_tca9535_expander(h, TCA9535ADDR_RS, ( ((lp-1)&0x7) < 4)?2:3, data, mask))
    goto error;

  /* detach i2c line back */
  set_pca9544_multiplexer(h, PCA9544ADDR_RS(lp), 0);

  closei2c(h);

  return 0;

error:
  if (h->fd >= 0) {
    set_pca9544_multiplexer(h, PCA9544ADDR_RS(lp), 0);
    closei2c(h);
  }

  return -1;
}


int x10i2c_get_port_rs_outs(x10i2c_t *h, uint8_t lp, uint8_t mask, uint8_t *data)
{
  X10I2CDBG("prt: %02d mask: %02X \n",
             (uint16_t)lp, (uint16_t)mask);

  if (openi2c(h, X10I2C_MAIN_BUS))
    goto error;

  /* setup multiplexer to attach nessessary line */
  if (set_pca9544_multiplexer(h, PCA9544ADDR_RS(lp), PCA9544VALUE_EXP_RS))
    goto error;

  /* write to output registers 2 or 3 according to port number */
  if (get_tca9535_expander(h, TCA9535ADDR_RS, (((lp-1)&0x7) < 4)?0:1, data))
    goto error;

  (*data) &= mask;

  X10I2CDBG("prt: %02d mask: %02X result: %02X \n",
             (uint16_t)lp, (uint16_t)mask, (uint16_t)(*data));

  /* detach i2c line back */
  set_pca9544_multiplexer(h, PCA9544ADDR_RS(lp), 0);

  closei2c(h);

  return 0;

error:
  if (h->fd >= 0) {
    set_pca9544_multiplexer(h, PCA9544ADDR_RS(lp), 0);
    closei2c(h);
  }

  return -1;
}

int x10i2c_set_port_tx(x10i2c_t *h, uint8_t lp, uint8_t enable)
{
  return x10i2c_set_port_outs(h, lp, (enable)?0:X10_TCA9535PIN_TX_DIS,
                              X10_TCA9535PIN_TX_DIS);
}

int x10i2c_get_port_tx(x10i2c_t *h, uint8_t lp, uint8_t *enable)
{
  return x10i2c_get_port_inputs(h, lp, X10_TCA9535PIN_TX_DIS, enable);
}

int x10i2c_get_port_rxloss(x10i2c_t *h, uint8_t lp, uint8_t *enable)
{
  int r;
  r = x10i2c_get_port_inputs(h, lp, X10_TCA9535PIN_RX_LOS, enable);
  (*enable) = ! (*enable);
  return r;
}

int x10i2c_get_port_inputs2(x10i2c_t *h, uint8_t lp, uint8_t *inputs)
{
  int r;
  r = x10i2c_get_port_inputs(h, lp, X10_TCA9535PIN_RX_LOS|X10_TCA9535PIN_MOD_ABS|X10_TCA9535PIN_TX_FLT, inputs);
  (*inputs) = ! (*inputs);
  return r;
}

int x10i2c_set_port_linkled(x10i2c_t *h, uint8_t lp, uint8_t enable)
{
  return x10i2c_set_port_outs(h, lp, (enable)?0:X10_TCA9535PIN_LINK_LED,
                              X10_TCA9535PIN_LINK_LED);
}

int x10i2c_set_port_actled(x10i2c_t *h, uint8_t lp, uint8_t enable)
{
  return x10i2c_set_port_outs(h, lp, (enable)?0:X10_TCA9535PIN_ACT_LED,
                              X10_TCA9535PIN_ACT_LED);
}

int x10i2c_set_port_almled(x10i2c_t *h, uint8_t lp, uint8_t enable)
{
  return x10i2c_set_port_outs(h, lp, (enable)?0:X10_TCA9535PIN_ALM_LED,
                              X10_TCA9535PIN_ALM_LED);
}

int x10i2c_set_port_leds(x10i2c_t *h, uint8_t lp, uint8_t leds)
{
  if (h->x10model<=X10MODEL_SITARA)
    return x10i2c_set_port_outs(h, lp, ~leds, X10_TCA9535PIN_ALM_LED|X10_TCA9535PIN_ACT_LED|X10_TCA9535PIN_LINK_LED);

  if (h->x10model>=X10MODEL_SITARA2)
    return x10i2c_set_pca9552leds(h, lp, leds);

  return -1;
}

int set_pca9552_reg(x10i2c_t *h, uint8_t reg, uint8_t val)
{
  X10I2CDBG("reg: %02X val: %02X\n",
             (uint16_t)reg, (uint16_t)val);

  /* set slave address */
  if (set_slave_addr(h, PCA9552_ADDR))
   return -1;

  if (i2c_smbus_write_byte_data(h->fd, reg, val)<0) {
	X10I2CDBG("Can not write to i2c device 0x%02x: %s\n",
                (uint16_t) PCA9552_ADDR, strerror(errno));
    return -1;
  }

  return 0;
}

int get_pca9552_reg(x10i2c_t *h, uint8_t reg, uint8_t *val)
{
  int old;

  X10I2CDBG("reg: %02X\n",
             (uint16_t)reg);

  /* set slave address */
  if (set_slave_addr(h, PCA9552_ADDR))
   return -1;

  /* read old data */
  old = i2c_smbus_read_byte_data(h->fd, reg);
  if (old < 0) {
    X10I2CERR("%s: Can not read old data\n",__FUNCTION__);
    return -1;
  }

  (*val) = old;

  return 0;
}

uint8_t pca9552_initable[]={
    PCA9552_REG_PSC0, 0x40, /* Period */
    PCA9552_REG_PWM0, 0x80, /* Duty circle */
    PCA9552_REG_PSC1, 0,
    PCA9552_REG_PWM1, 0xE0,
    /* All LEDS off */
    PCA9552_REG_LS0, 0x55,
    PCA9552_REG_LS0+1, 0x55,
    PCA9552_REG_LS0+2, 0x55,
    PCA9552_REG_LS0+3, 0x55
};

int x10i2c_init_pca9552(x10i2c_t *h)
{
  int r = 0;
  unsigned int i;
  for (i = 0; i<sizeof(pca9552_initable); i += 2) {
    r |= set_pca9552_reg(h, pca9552_initable[i], pca9552_initable[i+1]);
  }
  return r;

#if 0
  r|=set_pca9552_reg(h, PCA9552_REG_PSC0, 0x40); /* Period */
  r|=set_pca9552_reg(h, PCA9552_REG_PWM0, 0x80); /* Duty circle */
  r|=set_pca9552_reg(h, PCA9552_REG_PSC1, 0);
  r|=set_pca9552_reg(h, PCA9552_REG_PWM1, 0xE0);
  /* All LEDS off */
  r|=set_pca9552_reg(h, PCA9552_REG_LS0, 0x55);
  r|=set_pca9552_reg(h, PCA9552_REG_LS0+1, 0x55);
  r|=set_pca9552_reg(h, PCA9552_REG_LS0+2, 0x55);
  r|=set_pca9552_reg(h, PCA9552_REG_LS0+3, 0x55);

  return 0;
#endif
}

int x10i2c_set_pca9552leds(x10i2c_t *h, uint8_t lp, uint8_t leds)
{
  uint8_t nv=0;
  X10I2CDBG("prt: %02d leds: %02X \n",
             (uint16_t)lp, (uint16_t)leds);

  if (openi2c(h, X10I2C_MAIN_BUS))
    goto error;

  /* setup multiplexer to attach nessessary line */
  if (set_pca9544_multiplexer(h, PCA9544ADDR(lp), PCA9544VALUE_PCA9552LEDS))
    goto error;
/*
  if (x10i2c_set_pca9552_reg(h, PCA9552_REG_LS0+((lp-1)&0x3), &ov))
    goto error;
*/

  /* value of LED select register
   * 0 -on
   * 1 - off
   * 2 - PWM0
   * 3 - PWM1
  */

  get_pca9552_reg(h, PCA9552_REG_LS0+((lp-1)&0x3), &nv);
  nv &= (~3);
  nv |= ((leds&X10_TCA9535PIN_LINK_LED)?0:1);
  nv &= (~(3<<2));
  nv |= (1<<2);
  set_pca9552_reg(h, PCA9552_REG_LS0+((lp-1)&0x3), nv);

  if (!(lp&1))
    lp--;
  else
    lp++;

  get_pca9552_reg(h, PCA9552_REG_LS0+((lp-1)&0x3), &nv);
  nv &= ~(0xF<<4);
  if (leds&X10_TCA9535PIN_ALM_LED) {
    nv|=1<<4;   /* green off */
  } else {
    if (leds&(X10_TCA9535PIN_ACT_LED|X10_PORTLED_ACT_NOBLINK))  {
      if (leds&X10_TCA9535PIN_ACT_LED) {
        nv|=(2<<4);  /* green blink */
        nv|=(2<<6);  /* red blink */
      }
    } else {
      nv|=(1<<4);  /* green off */
      nv|=(1<<6);  /* red off */
    }

  }
  set_pca9552_reg(h, PCA9552_REG_LS0+((lp-1)&0x3), nv);


  /* detach i2c line back */
  set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);

  closei2c(h);

  return 0;

error:
  if (h->fd >= 0) {
    set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);
    closei2c(h);
  }

  return -1;
}

int x10i2c_get_port_sfp_status(x10i2c_t *h, uint8_t lp, uint8_t *enable)
{
  return x10i2c_get_port_inputs(h, lp, X10_TCA9535PIN_MOD_ABS, enable);
}

int x10_test_ports_leds(x10i2c_t *h, uint16_t timer)
{
  uint8_t p;

  for (p = 1; p <= 24; ++p)
    x10i2c_set_port_leds(h, p, X10_TCA9535PIN_LINK_LED);
  usleep(1000*timer);

  for (p = 1; p <= 24; ++p)
    x10i2c_set_port_leds(h, p, (h->x10model>2)?X10_PORTLED_ACT_NOBLINK:X10_TCA9535PIN_ACT_LED);

  usleep(1000*timer);

  for (p = 1; p <= 24; ++p)
    x10i2c_set_port_leds(h, p, X10_TCA9535PIN_ALM_LED);

  usleep(1000*timer);
  for (p = 1; p <= 24; ++p)
    x10i2c_set_port_leds(h, p, 0);

  return 0;
}


int x10_set_port_rs0_out(x10i2c_t *h, uint8_t lp, uint8_t enable)
{  
  uint8_t shift;
  if (h->x10model<=X10MODEL_SITARA) {
   shift = (((lp-1)&0x03)<<1);
   return x10i2c_set_port_rs_outs(h, lp, ((enable)?1:0) << shift, 0x1 << shift);
  } else
   return x10i2c_set_port_outs(h, lp, (enable)?X10_TCA9535PIN_RS0:0, X10_TCA9535PIN_RS0);
}

int x10_set_port_rs1_out(x10i2c_t *h, uint8_t lp, uint8_t enable)
{
  uint8_t shift;
  if (h->x10model<=X10MODEL_SITARA) {
    shift = (((lp-1)&0x03)<<1);
    return x10i2c_set_port_rs_outs(h, lp, ((enable)?2:0) << shift, 0x2 << shift);
  } else {
    return x10i2c_set_port_outs(h, lp, (enable)?X10_TCA9535PIN_RS1:0, X10_TCA9535PIN_RS1);
  }
}

int x10_get_port_rs0_out(x10i2c_t *h, uint8_t lp, uint8_t *enable)
{
  uint8_t shift;
  if (h->x10model<=X10MODEL_SITARA) {
   shift = (((lp-1)&0x03)<<1);
   return x10i2c_get_port_rs_outs(h, lp, 0x1 << shift, enable);
  } else
   return x10i2c_get_port_outs(h, lp, X10_TCA9535PIN_RS0, enable);
}

int x10_get_port_rs1_out(x10i2c_t *h, uint8_t lp, uint8_t *enable)
{
  uint8_t shift;
  if (h->x10model<=X10MODEL_SITARA) {
   shift = (((lp-1)&0x03)<<1);
   return x10i2c_get_port_rs_outs(h, lp, 0x2 << shift, enable);
  } else
   return x10i2c_get_port_outs(h, lp, X10_TCA9535PIN_RS1, enable);
}

int x10_get_port_sfp_pwrok(x10i2c_t *h, uint8_t lp, uint8_t *enable)
{
  if (h->x10model==X10MODEL_SITARA2)
    return x10i2c_get_port_inputs(h, lp, X10_TCA9535PIN_PWOK, enable);
  (*enable)=1;
  return -1;
}

int x10_set_port_sfp_pins(x10i2c_t *h, uint8_t lp, uint16_t pins)
{
  return 1;
}

int x10_get_port_sfp_pins(x10i2c_t *h, uint8_t lp, uint16_t *pins)
{
  return 1;
}

#define SFPREG_VENDOR       20
#define SFPREG_PARTNUMBER   40
#define SFPREG_TX_POWER_MSB 102
#define SFPREG_TX_POWER_LSB 103
#define SFPREG_RX_POWER_MSB 104
#define SFPREG_RX_POWER_LSB 105

int x10i2c_get_sfp_regs(x10i2c_t *h, uint8_t lp, uint8_t i2caddr, uint8_t reg, uint16_t count, void *data)
{
  int res, c;

  X10I2CDBG("prt: %02d reg: %03d cnt: %d\n",
             (uint16_t)lp, (uint16_t)reg,
             (uint16_t)count);

  if (openi2c(h, X10I2C_MAIN_BUS))
    goto error;

  /* setup multiplexer to attach nessessary line */
  if (set_pca9544_multiplexer(h, PCA9544ADDR(lp), PCA9544VALUE_SFP(lp)))
    goto error;

  /* set slave address */
  if (set_slave_addr(h, i2caddr))
    goto error;

  /* read registers data */
  for (c=0; c<count;++c) {
    res = i2c_smbus_read_byte_data(h->fd, reg+c);

    if (res<0) {
      X10I2CERR("Can not read from i2c device 0x%02x: %s\n",
                i2caddr, strerror(errno));
      goto error;
    }
    ((uint8_t*)data)[c] = (uint8_t)res;
  }

  /* detach i2c line back */
  set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);

  closei2c(h);

  return 0;

error:

  if (h->fd >= 0) {
    set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);
    closei2c(h);
  }

  return -1;
}


int x10i2c_set_sfp_regs(x10i2c_t *h, uint8_t lp, uint8_t i2caddr, uint8_t reg, uint16_t count, void *data)
{
  int res, c;

  X10I2CDBG("prt: %02d reg: %03d cnt: %d\n",
             (uint16_t)lp, (uint16_t)reg,
             (uint16_t)count);

  if (openi2c(h, X10I2C_MAIN_BUS))
    goto error;

  /* setup multiplexer to attach nessessary line */
  if (set_pca9544_multiplexer( h, PCA9544ADDR(lp), PCA9544VALUE_SFP(lp)) )
    goto error;

  /* set slave address */
  if (set_slave_addr( h, i2caddr) )
    goto error;

  /* read registers data */
  for ( c = 0; c < count; ++c ) {
    res = i2c_smbus_write_byte_data(h->fd, reg+c, ((uint8_t*)data)[c]);

    if (res<0) {
      X10I2CERR("Can not write to i2c device 0x%02x: %s\n",
                i2caddr, strerror(errno));
      goto error;
    }
  }

  /* detach i2c line back */
  set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);

  closei2c(h);

  return 0;

error:
  if (h->fd >= 0) {
    set_pca9544_multiplexer( h, PCA9544ADDR(lp), 0 );
    closei2c(h);
  }

  return -1;
}

int x10_echo(char *value, char *filename)
{
  FILE *f;
  int l;

  f = fopen(filename, "wt");
  if (!f)
    return -1;

  l = strlen(value);
  if (fwrite(value, l, 1, f) != 1) {
    return -1;
  }

  return fclose(f);
}

#define X10_ECHO(v, f) if (x10_echo(v, f)) \
{ \
  X10I2CERR("Can not write '%s' to '%s' - %s\n", v, f, strerror(errno));  \
  return -1; \
}



#define PWMCOOLERN1 1
#define PWMCOOLERN2 0
char pwmcoolerprefix[]="/sys/class/pwm/ehrpwm";

char pwmrequest[]="request";
char pwmperiod_freq[]="period_freq";
char pwmduty_percent[]="duty_percent";
char pwmrun[]="run";
#define ACCESS_RIGHTS S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH

int x10_init_gpio_out(x10i2c_t *h, uint8_t pin)
{
   struct stat st;
   snprintf(h->s, sizeof(h->s), "/sys/class/gpio/gpio%u", (uint16_t)pin);
   if (stat(h->s, &st)) {
     snprintf(h->s, sizeof(h->s), "%u", (uint16_t)pin);
     X10_ECHO(h->s, "/sys/class/gpio/export");
   }

   snprintf(h->s, sizeof(h->s), "/sys/class/gpio/gpio%u/direction", (uint16_t)pin);
   x10_echo("out",h->s);

   snprintf(h->s, sizeof(h->s), "/sys/class/gpio/gpio%u/value", (uint16_t)pin);
   chmod(h->s, ACCESS_RIGHTS);
   return 0;
}

int x10_set_gpio_pin(x10i2c_t *h, uint8_t pin,  uint8_t v)
{
  snprintf(h->s, sizeof(h->s), "/sys/class/gpio/gpio%u/value", (uint16_t)pin);
  X10_ECHO((v)?"1":"0", h->s);
  return 0;
}

int x10_init_gpio(x10i2c_t *h)
{
  if (x10_init_gpio_out(h, (h->x10model>=X10MODEL_SITARA)?GPIO_RESET_SITARA:GPIO_RESET))
    return -1;

  /* enable chassis fans on SITARA model */
  if (h->x10model==X10MODEL_SITARA) {
    if (x10_init_gpio_out(h, 46))
      return -1;
    if (x10_set_gpio_pin(h, 46, 1))
      return -1;
  }

  /* setup MUX mode to GPIO */
  if ((h->x10model==X10MODEL_BBONE))
   X10_ECHO("7","/sys/kernel/debug/omap_mux/gpmc_ad7");

  if (x10_init_gpio_out(h, (h->x10model>=X10MODEL_SITARA2)?GPIO_POWERLED_SITARA2:GPIO_POWERLED))
    return -1;

  if ((h->x10model>=X10MODEL_SITARA2)) {
    if (x10_init_gpio_out(h, GPIO_EPROMWP_SITARA2))
       return -1;
  }

  return 0;
}

#define BUILDCOOLERSPWMFNAME(file) \
{ \
  snprintf(h->s, sizeof(h->s), "%s.%u:%u/%s", pwmcoolerprefix, PWMCOOLERN1, PWMCOOLERN2, file); \
}

int x10_init_pwm(x10i2c_t *h)
{
  /* setup MUX mode to PWM */
  X10_ECHO("6","/sys/kernel/debug/omap_mux/gpmc_a2");

  BUILDCOOLERSPWMFNAME(pwmrequest);
  chmod(h->s, ACCESS_RIGHTS);
  BUILDCOOLERSPWMFNAME(pwmperiod_freq);
  chmod(h->s, ACCESS_RIGHTS);
  BUILDCOOLERSPWMFNAME(pwmduty_percent);
  chmod(h->s, ACCESS_RIGHTS);
  BUILDCOOLERSPWMFNAME(pwmrun);
  chmod(h->s, ACCESS_RIGHTS);

  /* disable before */
  BUILDCOOLERSPWMFNAME(pwmrun);
  x10_echo("0", h->s);
  BUILDCOOLERSPWMFNAME(pwmrequest);
  x10_echo("0", h->s);

  /* init PWM parameters */
  BUILDCOOLERSPWMFNAME(pwmrequest);
  X10_ECHO("1", h->s);
  BUILDCOOLERSPWMFNAME(pwmperiod_freq);
  X10_ECHO("100", h->s);
  BUILDCOOLERSPWMFNAME(pwmduty_percent);
  X10_ECHO("0", h->s);
  BUILDCOOLERSPWMFNAME(pwmrun);
  X10_ECHO("1", h->s);

  return 0;
}


int x10_reset_pin(x10i2c_t *h, uint8_t v)
{
  return x10_set_gpio_pin(h, (h->x10model>=X10MODEL_SITARA)?GPIO_RESET_SITARA:GPIO_RESET, v);
}

int x10_powerled_pin(x10i2c_t *h, uint8_t v)
{
  if (h->x10model>=X10MODEL_SITARA2)
    v=!v;
  return x10_set_gpio_pin(h, (h->x10model>=X10MODEL_SITARA2)?GPIO_POWERLED_SITARA2:GPIO_POWERLED, v);
}

int x10_reset(x10i2c_t *h)
{
  if (x10_set_gpio_pin(h, (h->x10model>=X10MODEL_SITARA)?GPIO_RESET_SITARA:GPIO_RESET, 1))
    return -1;
  usleep(100000);
  if (x10_set_gpio_pin(h, (h->x10model>=X10MODEL_SITARA)?GPIO_RESET_SITARA:GPIO_RESET, 0))
    return -1;
  return 0;
}

int x10_set_coolers(x10i2c_t *h, uint8_t v)
{
  char s[6];

  if (h->x10model>=X10MODEL_SITARA)
    return 0;

  if (v != 0)
   v=(80*v)/100+20;

  sprintf(s, "%u", v);

  if (h->x10model==X10MODEL_BBONE) {
    BUILDCOOLERSPWMFNAME(pwmduty_percent);
    X10_ECHO(s, h->s);
  }

  return 0;
}

int x10_get_ads7830_reg(x10i2c_t *h, uint8_t i2caddr,  uint8_t reg, uint8_t *temp)
{
  int r;

  if (h->x10model<=X10MODEL_SITARA) {
    if (openi2c(h, X10I2C_MAIN_BUS))
      goto error;
  } else {
    if (openi2c(h, X10I2C_SITARA_BUS))
      goto error;
    change_sitara_i2c_bus(h, 0);
  }


  /* set slave address */
  if (set_slave_addr( h, i2caddr) )
    goto error;

  /* read registers data */
  r = i2c_smbus_read_byte_data(h->fd, reg);

  X10I2CDBG("reg: %02X val: %02X\n",
              (uint16_t)reg, r);

  if (r < 0)
    goto error;

  if (h->x10model>=X10MODEL_SITARA2)
    restore_sitara_i2c_bus(h);
  closei2c(h);

  (*temp) = (uint8_t)r;
  return 0;

error:

  if (h->x10model>=X10MODEL_SITARA2)
    restore_sitara_i2c_bus(h);
  closei2c(h);
  return -1;
}

#define TEMP_COUNT 76
int tempr_to_code[] = {188,186,184,182,179,177,174,172,170,
                       167,165,162,160,157,155,152,150,147,
                       145,142,140,137,135,132,130,128,125,
                       123,120,118,116,113,111,109,106,104,
                       102,100,98,96,94,92,90,88,86,84,
                       82,80,78,77,75,73,72,70,68,67,
                       65,64,62,61,59,58,57,55,54,53,
                       52,50,49,48,47,46,45,44,43,42};


int tempr_to_code_sitara2[] = {250,247,244,241,237,234,231,228,225,
                       222,218,215,212,208,205,202,199,195,
                       192,189,185,182,179,175,172,169,166,
                       163,159,156,153,150,147,144,141,138,
                       135,132,130,127,124,121,119,116,114,
                       111,109,106,104,102,99,97,95,93,90,
                       88,86,84,82,81,79,77,75,73,72,70,68,
                       67,65,64,62,61,59,58,57,55};

int x10_decode_temp(x10i2c_t *h, int *table, int code, uint8_t *t)
{
  uint8_t i;

  if (code > table[0]) return -1; // error of sensor
  if (code < table[TEMP_COUNT-1])
    (*t) = 76; // max temperature

  for (i = 0; i < TEMP_COUNT; i++)
    if (code >= table[i]) {
      (*t) = i;
      return 0;
    }

  X10I2CDBGL(2, "%s: can't decode temp code %i\n",  code);

  return -1;
}

int x10_get_env_temp(x10i2c_t *h, uint8_t *temp)
{  
  uint8_t code;
  uint16_t t[6], max;
  if (h->x10model<=X10MODEL_SITARA) {
    if (x10_get_ads7830_reg(h, ADS7830ADDR, 0xCC, &code))
      return -1;
    return x10_decode_temp(h, tempr_to_code, code, temp);
  } else {
    if (x10i2c_get_sitara2_temps(h, t))
      return -1;

    max = 0;
    if (max<t[2]) max = t[2];
    if (max<t[3]) max = t[3];
    if (max<t[5]) max = t[5];

    (*temp) = (max>>2)&0xFF;
    return 0;
  }
}

int x10_get_chip_temp(x10i2c_t *h, uint8_t *temp)
{
  uint8_t code;
  uint16_t t[6];
  if (h->x10model<=X10MODEL_SITARA) {
    if (x10_get_ads7830_reg(h, ADS7830ADDR, 0x8C, &code))
      return -1;
  } else {
    if (x10i2c_get_sitara2_temps(h, t))
      return -1;
    (*temp) = (t[0]>>2)&0xFF;
    return 0;
  }

  return x10_decode_temp(h, tempr_to_code, code, temp);
}

int read_eeprom_block(x10i2c_t *h, uint8_t i2caddr, uint16_t memaddr, uint16_t count, void *data)
{
  int r;
  uint8_t buf[3];
  X10I2CDBG(" addr: %X cnt: %u\n",
              memaddr, count);

  if (!count)
    return 0;

  /* set slave address */
  if (set_slave_addr( h, i2caddr) )
    goto error;

  buf[0]=(memaddr>>8);
  buf[1]=memaddr&0xFF;

  r =write(h->fd, buf, 2);
  if (r!=2) {
    X10I2CERR("%s: write error1\n", __FUNCTION__);
    goto error;
  }

  if (read(h->fd, data, 1)!=1) {
    X10I2CERR("%s: read error1\n", __FUNCTION__);
    goto error;
  }

  if (count>=1) {

   /* set slave address */
   if (set_slave_addr( h, i2caddr) )
     goto error;

   r= read(h->fd, &((uint8_t*)data)[1], (count-1));
   if (r!=(count-1)) {
     X10I2CERR("%s: read2 r=%i \n",  r);
     goto error;
   }

  }

  return 0;

error:
  return -1;
}

#define EEPROM_BLSIZE 0x2000

int read_eeprom(x10i2c_t *h, uint8_t i2caddr, uint16_t saddr, uint16_t count, void *data)
{
  uint16_t size, maddr;

  if (!count)
    return 0;

  maddr=saddr;
  while (count>0) {
    size=(count>EEPROM_BLSIZE)?EEPROM_BLSIZE:count;
    if (read_eeprom_block(h, i2caddr, maddr, size, &((uint8_t*)data)[maddr-saddr]))
      return -1;
    count-=size;
    maddr+=size;
  }
  return 0;
}

int x10_read_eeprom(x10i2c_t *h, uint16_t memaddr, uint16_t count, void *data)
{
    X10I2CDBG(" addr: %X cnt: %u\n",
                memaddr, count);
    if (!count)
      return 0;

    if (((uint32_t)memaddr+(uint32_t)count-1)>0xFFFF)
      return -1;

    if (openi2c(h, X10I2C_SITARA_BUS))
       goto error;

    if (h->x10model>=2)
     change_sitara_i2c_bus(h, 0);

    if (read_eeprom(h, SITARA_I2C_EEPROM, memaddr, count, data))
      goto error;

    if (h->x10model>=2)
      restore_sitara_i2c_bus(h);
    closei2c(h);

    return 0;

error:

    if (h->x10model>=2)
      restore_sitara_i2c_bus(h);
    /*x10_reset_pin(h, 0);*/
    closei2c(h);

    return -1;
}


int set_sitara_eeprom_protect(x10i2c_t *h, char protect)
{
  return x10_set_gpio_pin(h, GPIO_EPROMWP_SITARA2, (protect)?1:0);
}

int write_eeprom(x10i2c_t *h, uint8_t i2caddr, uint16_t memaddr, uint16_t count, void *data)
{
  int c;
  uint8_t buf[3];

  X10I2CDBG(" addr: %X cnt: %u\n",
              memaddr, count);

  for (c=0; c<count; ++c, ++memaddr) {

  /* set slave address */      
  if (set_slave_addr( h, i2caddr))
    goto error;

  X10I2CDBGL(3, "%s: write %u %02X\n",  c, ((uint8_t*)data)[c]);
  /* set address */
  buf[0]=(memaddr>>8);
  buf[1]=memaddr&0xFF;
  buf[2]=((uint8_t*)data)[c];

  if (write(h->fd, buf, 3)!=3) {
    X10I2CERR("%s: write error1\n", __FUNCTION__);
    goto error;
  }
  usleep(5000);
  }

  return 0;

error:
  return -1;

}

int x10_write_eeprom(x10i2c_t *h, uint16_t memaddr, uint16_t count, void *data)
{
  X10I2CDBG(" addr: %X cnt: %u\n",
              memaddr, count);

  if (!count)
    return 0;

  if (((uint32_t)memaddr+(uint32_t)count-1)>0xFFFF)
    return -1;

  /*x10_reset_pin(h, 1);*/
  if (h->x10model>=2)
    set_sitara_eeprom_protect(h, 0);

  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  if  (h->x10model>=2)
    change_sitara_i2c_bus(h, 0);

  if (write_eeprom(h, SITARA_I2C_EEPROM, memaddr, count, data))
    goto error;

  /*x10_reset_pin(h, 0);*/
  if  (h->x10model>=2) {
    restore_sitara_i2c_bus(h);
    set_sitara_eeprom_protect(h, 1);
  }
  closei2c(h);

  return 0;

error:

  /*x10_reset_pin(h, 0);*/
  if  (h->x10model>2) {
    restore_sitara_i2c_bus(h);
    set_sitara_eeprom_protect(h, 1);
  }
  closei2c(h);

  return -1;
}

uint8_t xor_mem(void *buf, size_t size)
{
  size_t i;
  uint8_t r;
  r = ((uint8_t*)buf)[0];
  for (i=1;i<size;++i)
    r ^= (((uint8_t*)buf)[i]);
  return r;
}

unsigned short crc16b( const void *data, unsigned short size, unsigned short crc )
{
  char i;
  const unsigned char *ptr = ( const unsigned char * ) data;
  while( size-- > 0 ) {
    crc = crc ^ ( unsigned short ) *ptr++ << 8;
    i = 8;
    do {
      if( crc & 0x8000 )
        crc = crc << 1 ^ 0x1021;
      else
        crc = crc << 1;
    } while( --i );
  }
  return crc;
}

int x10_update_eeprom(x10i2c_t *h, FILE *out, char *fname, x10_eeprom_info_t *info)
{
  uint8_t *buf;
  FILE *file;
  struct stat s;
  uint32_t maddr, size, cnt, serial;
  x10_eeprom_info_t iinfo, *inf;

  X10I2CDBG(" file: %s\n",
              fname);

  if (stat(fname, &s))
    return -1;  

  if ( (unsigned int)s.st_size > (0xFFFF-sizeof(x10_eeprom_info_t)) ) {
    X10I2CERR("%s: file size too big %u\n", s.st_size);
    return -1;
  }

  buf=malloc(s.st_size);
  if (!buf)
    return -1;

  file = fopen(fname, "rb+");
  if (!file) {
    X10I2CERR("%s: can't open file %s\n", fname);
    goto error;
  }

  if ( fread(buf, 1, s.st_size, file) != (unsigned int)s.st_size) {
    X10I2CERR("%s: can't read %u from file %s\n", s.st_size, fname);
    goto error;
  }

  fclose(file);
  file=0;

  if (out) {
    maddr=0;
    cnt=s.st_size;
    while (cnt>0) {
       size=(cnt<512)?cnt:512;
       if (x10_write_eeprom(h, maddr, size, &buf[maddr]))
         goto error;       
       maddr+=size;
       cnt-=size;
       fprintf(out, "Writed %u bytes %lu%%\n", maddr, (maddr)*100/s.st_size);
    }
  } else
    if (x10_write_eeprom(h, 0, s.st_size, buf))
      goto error;

  serial = 0;
  if (!x10_get_eeprom_info(h, &iinfo))
    serial = iinfo.serial;
  else
    memset(&iinfo, 0 ,sizeof(x10_eeprom_info_t));

  inf = info;
  if (!inf) {
   inf = &iinfo;
   inf->ver = 0;
   inf->serial = 0;
  }

  if (serial)
    inf->serial = serial;

  struct tm *t;
  struct timeval tv;
  gettimeofday(&tv, 0);
  t = localtime(&tv.tv_sec);
  inf->year = t->tm_year+1900;
  inf->month = t->tm_mon;
  inf->day = t->tm_mday;

  inf->sum=crc16b(buf, s.st_size, 0);
  inf->size=s.st_size;
  inf->hsum = xor_mem(inf, sizeof(x10_eeprom_info_t)-1);
  inf->hsum = ~inf->hsum;

  X10I2CDBG("writing eeprom header hsum %02X date %04u/%02u/%02u...\n",
             inf->hsum, inf->year, inf->month, inf->day);

  if (x10_write_eeprom(h, 0xFFFF-sizeof(x10_eeprom_info_t)+1, sizeof(x10_eeprom_info_t), inf))
    goto error;

  free(buf);

  return 0;

error:
  if (file)
    fclose(file);

  if (buf)
    free(buf);

  return -1;
}

int x10_get_eeprom_info(x10i2c_t *h, x10_eeprom_info_t *info)
{
  X10I2CDBG("\n");
  uint8_t sum;

  if (x10_read_eeprom(h, 0xFFFF-sizeof(x10_eeprom_info_t)+1, sizeof(x10_eeprom_info_t), info))
    return -1;

  sum=(~xor_mem(info, sizeof(x10_eeprom_info_t)-1));
  if (info->hsum != sum) {
    X10I2CERR("checksum wrong %X != %X\n", info->hsum, sum);
    return -1;
  }

  return 0;
}


int x10_check_eeprom(x10i2c_t *h, uint8_t *ok)
{    
  x10_eeprom_info_t info;
  void *buf=0;

  X10I2CDBG("%s:\n", __FUNCTION__);
  (*ok)=0;

  if (x10_get_eeprom_info(h, &info))
    return -1;

  buf=malloc(info.size);
  if (!buf)
    return -1;

  if (x10_read_eeprom(h, 0, info.size, buf))
    goto error;

  if (info.sum != crc16b(buf, info.size, 0)) {
    X10I2CERR("%s: wrong checksum\n",__FUNCTION__);
    goto error;
  }  

  (*ok)=1;
  free(buf);

  return 0;

error:
  if (buf)
    free(buf);

  return -1;
}

int x10i2c_fujitsu_set_mainreg(x10i2c_t *h, uint8_t i2caddr, uint16_t addr, uint32_t value)
{
  unsigned char buf[6];

  X10I2CDBG(" addr: %X value: %08X\n",
              addr, value);

  /* set slave address */
  if (set_slave_addr( h, i2caddr))
    return -1;

  buf[0] = (addr & 0xFF);
  buf[1] = (addr >> 8);

  buf[2] = (value & 0xFF);
  buf[3] = ((value>>8) & 0xFF);
  buf[4] = ((value>>16) & 0xFF);
  buf[5] = ((value>>24) & 0xFF);

  if (write(h->fd, buf, 6)!=6) {
    X10I2CERR("%s: write error: %s\n",
                strerror(errno));
    return -1;
  }

  return 0;
}

inline int x10i2c_fujitsu_set_reg(x10i2c_t *h, uint16_t addr, uint32_t value)
{

  X10I2CDBG(" addr: %X value: %08X\n",
              addr, value);

  if (h->x10model < X10MODEL_SITARA)
    return -1;

  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  change_sitara_i2c_bus(h, 1);

  if (x10i2c_fujitsu_set_mainreg(h, SITARA_I2C_FUJITSU , addr, value))
    goto error;

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return 0;

error:

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return -1;
}

int x10i2c_fujitsu_get_mainreg(x10i2c_t *h, uint8_t i2caddr, uint16_t addr, uint32_t *value)
{
  unsigned char buf[4];

  X10I2CDBG(" addr: %X \n",
              addr);

  /* set slave address */
  if (set_slave_addr( h, i2caddr))
    return -1;

  buf[0] = (addr & 0xFF);
  buf[1] = (addr >> 8);  

  if (write(h->fd, buf, 2)!=2) {
    X10I2CERR("%s: write error: %s\n",
                strerror(errno));
    return -1;
  }

  if (read(h->fd, buf, 4)!=4) {
    X10I2CERR("%s: read error: %s\n",
                strerror(errno));
    return -1;
  }

  (*value) = buf[0];
  (*value) |= (buf[1]<<8);
  (*value) |= (buf[2]<<16);
  (*value) |= (buf[3]<<24);

  X10I2CDBGL(2, "%s: value: %08X\n",
               (*value));

  return 0;

}

inline int x10i2c_fujitsu_get_reg(x10i2c_t *h, uint16_t addr, uint32_t *value)
{
  X10I2CDBG(" addr: %X \n",
              addr);

  if (h->x10model < X10MODEL_SITARA)
      return -1;

  if (openi2c(h, X10I2C_SITARA_BUS))
      goto error;

  change_sitara_i2c_bus(h, 1);

  if (x10i2c_fujitsu_get_mainreg(h, SITARA_I2C_FUJITSU , addr, value))
    goto error;

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return 0;

error:

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return -1;
}

int x10i2c_fujitsu_set_port_reg(x10i2c_t *h, uint16_t pp, uint16_t addr, uint32_t value)
{
  if (x10i2c_fujitsu_set_reg(h, 0x10, 0x4000 + 0x800*pp))
    return -1;
  return x10i2c_fujitsu_set_reg(h, addr + 0x3000,  value);
}

int x10i2c_fujitsu_get_port_reg(x10i2c_t *h, uint16_t pp, uint16_t addr, uint32_t *value)
{
  if (x10i2c_fujitsu_set_reg(h, 0x10, 0x4000 + 0x800*pp))
    return -1;
  return x10i2c_fujitsu_get_reg(h, addr + 0x3000,  value);
}


int x10i2c_set_reg(x10i2c_t *h, uint8_t i2caddr, uint8_t reg, uint8_t val)
{
  X10I2CDBG("dev: %02X reg: %02X val: %02X\n",
             (uint16_t)i2caddr, (uint16_t)reg, (uint16_t)val);

  /* set slave address */
  if (set_slave_addr(h, i2caddr))
   return -1;

  if (i2c_smbus_write_byte_data(h->fd, reg, val)<0) {
	X10I2CERR("Can not write to i2c device 0x%02x: %s\n",
                (uint16_t) i2caddr, strerror(errno));
    return -1;
  }

  return 0;
}

int x10i2c_get_reg(x10i2c_t *h, uint8_t i2caddr, uint8_t reg, uint8_t *val)
{
  int old;


  /* set slave address */
  if (set_slave_addr(h, i2caddr))
    return -1;

  /* read old data */
  old = i2c_smbus_read_byte_data(h->fd, reg);
  if (old < 0){
    X10I2CERR("%s: Can not read old data\n",__FUNCTION__);
    return -1;
  }  
  X10I2CDBG("data: %02X value: %02X\n",
              (uint16_t)reg, (uint16_t)old);
  (*val) =  old;

  return 0;
}

inline int x10i2c_get_ads7830_reg(x10i2c_t *h, uint8_t i2caddr,  uint8_t reg, uint8_t *val)
{
  return x10i2c_get_reg(h, i2caddr, reg, val);
}

int x10i2c_init_ADT7463(x10i2c_t *h, uint8_t *table, uint16_t tsize, uint8_t i2caddr)
{
  uint8_t i;
  X10I2CDBG("addr: %02X\n",(uint16_t)i2caddr);
  for (i=0; i<(tsize); i+=2)
    if (x10i2c_set_reg(h, i2caddr, table[i],
                                   table[i+1]))
      return -1;

  return 0;
}

int set_ADT7463_fan_speed(x10i2c_t *h, uint8_t i2caddr, uint8_t fan, uint8_t percent)
{
  uint8_t value;

  if (percent<=100)
   value = ((uint16_t)percent)*255/100;
  else
   value=0xFF;

  return x10i2c_set_reg(h, i2caddr, 0x30+fan, value);
}

int x10i2c_get_ADT7463_temp(x10i2c_t *h, uint8_t i2caddr, uint8_t tnum, uint16_t *value)
{
  uint8_t lsb,v;

  if (tnum>2)
    return -1;

  if (x10i2c_get_reg(h, i2caddr, 0x77, &lsb))
    return -1;

  if (x10i2c_get_reg(h, i2caddr, 0x25+tnum, &v))
    return -1;

  (*value) = ((lsb>>(2+(tnum<<1))) & 0x3) | (((uint16_t)v) << 2);

  return 0;
}

int x10i2c_get_ADT7463_temps(x10i2c_t *h, uint8_t i2caddr, uint16_t *temps)
{
  uint8_t lsb, v[3];

  if (x10i2c_get_reg(h, i2caddr, 0x77, &lsb))
    return -1;

  if (x10i2c_get_reg(h, i2caddr, 0x25, &v[0]))
    return -1;

  if (x10i2c_get_reg(h, i2caddr, 0x26, &v[1]))
    return -1;

  if (x10i2c_get_reg(h, i2caddr, 0x27, &v[2]))
    return -1;

  temps[0]= ((lsb>>2) & 0x3) | (((uint16_t)v[0]) << 2);
  temps[1]= ((lsb>>4) & 0x3) | (((uint16_t)v[1]) << 2);
  temps[2]= ((lsb>>6) & 0x3) | (((uint16_t)v[2]) << 2);

  return 0;
}

int x10i2c_get_sitara2_temps(x10i2c_t *h, uint16_t *temps)
{
  if (h->x10model < X10MODEL_SITARA)
    return -1;

  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  change_sitara_i2c_bus(h, 1);

  if (x10i2c_get_ADT7463_temps(h, SITARA_MAIN_ADT7463, temps))
    goto error;

  if (x10i2c_get_ADT7463_temps(h, SITARA_CHASSIS_ADT7463, &temps[3]))
    goto error;

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return 0;

error:

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return -1;
}

typedef struct x10_sitara2_temps_s {
  uint16_t chip;
  uint16_t sys[3];
  uint16_t pw[2];
} x10_sitara2_temps_t;

uint16_t ADT7463_temp2mc(uint16_t t)
{
  return ((t>>2)&0xFF)*1000+((t&0x3)*250);
}

int x10_conv_sitara2_temps(x10i2c_t *h, uint16_t *t, uint8_t *v, uint8_t *sio, x10_sitara2_temps_t *st)
{
  uint8_t tm;

  X10I2CDBG("%s:\n", __FUNCTION__);
  st->chip = ADT7463_temp2mc(t[0]);
  st->sys[0] = ADT7463_temp2mc(t[2]);
  st->sys[1] = ADT7463_temp2mc(t[3]);
  st->sys[2] = ADT7463_temp2mc(t[5]);

  if (sio[0]&SITARA_SYSIO_TCA9535_PSB0_PRESENT)
    if (x10_decode_temp(h, tempr_to_code_sitara2, v[0], &tm))
      return -1;
  st->pw[0]=tm*1000;

  if (sio[0]&SITARA_SYSIO_TCA9535_PSB1_PRESENT)
    if (x10_decode_temp(h, tempr_to_code_sitara2, v[1], &tm))
      return -1;

  st->pw[1]=tm*1000;

  return 0;
}

int x10i2c_get_ADT7463_fan(x10i2c_t *h, uint8_t i2caddr, uint8_t fnum, uint16_t *fan)
{
  uint8_t lsb, msb;

  if (x10i2c_get_reg(h, i2caddr, 0x28+(fnum<<1), &lsb))
    return -1;

  if (x10i2c_get_reg(h, i2caddr, 0x28+(fnum<<1)+1, &msb))
    return -1;

  fan[0] = ((uint16_t)lsb | (((uint16_t)msb)<<8));  

  if ((fan[0]<0xFFFF) && (fan[0]))
   fan[0]=(90000*60)/fan[0];
  else
   fan[0]=0;

  return 0;
}

int x10i2c_get_sitara2_fans(x10i2c_t *h, uint16_t *fans)
{
  uint8_t f;
  if (h->x10model < X10MODEL_SITARA2)
    return -1;

  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  change_sitara_i2c_bus(h, 1);

  for (f=0; f<4; ++f)
    if (x10i2c_get_ADT7463_fan(h, SITARA_MAIN_ADT7463, f, &fans[f]))
      goto error;

  for (f=0; f<4; ++f)
    if (x10i2c_get_ADT7463_fan(h, SITARA_CHASSIS_ADT7463, f, &fans[f+4]))
      goto error;

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return 0;

error:

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return -1;
}

int x10i2c_get_ads7830_voltage(x10i2c_t *h, uint8_t i2caddr, uint8_t ch, uint8_t *v)
{
  uint8_t r;
  r = (ch&1)?(4+(ch>>1)):(ch>>1);
  if (x10i2c_get_ads7830_reg(h, i2caddr, (0x8C)|((r&7)<<4), v))
      return -1;
  return 0;
}

int x10i2c_get_ads7830_voltages(x10i2c_t *h, uint8_t i2caddr, uint8_t *v)
{
  uint8_t ch, r;
  for (ch=0; ch<8; ++ch) {      
    r = (ch&1)?(4+(ch>>1)):(ch>>1);
    if (x10i2c_get_ads7830_reg(h, i2caddr, (0x8C)|((r&7)<<4), &v[ch]))
      return -1;
  }
  return 0;
}

int x10i2c_get_sitara2_voltages(x10i2c_t *h, uint8_t *v)
{
  if (h->x10model < X10MODEL_SITARA2)
    return -1;

  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  change_sitara_i2c_bus(h, 1);

  if (x10i2c_get_ads7830_voltages(h, SITARA_ADS7830ADDR1, v))
    return -1;

  if (x10i2c_get_ads7830_voltages(h, SITARA_ADS7830ADDR2, &v[8]))
    return -1;

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return 0;

error:

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return -1;
}

int x10i2c_get_sitara2_sio_values(x10i2c_t *h, uint8_t *sio)
{
  if (h->x10model < X10MODEL_SITARA2)
    return -1;

  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  change_sitara_i2c_bus(h, 1);

  if (get_tca9535_expander(h, SITARA_SYSIO_TCA9535ADDR, 0, sio))
    return -1;

  if (get_tca9535_expander(h, SITARA_SYSIO_TCA9535ADDR, 1, &sio[1]))
    return -1;

  X10I2CDBG("sio0 %02X sio1 %02X\n",  sio[0], sio[1]);

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return 0;

error:

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return -1;
}

typedef struct x10_sitara2_voltages_s {
    uint16_t pw0;
    uint16_t pw1;
    uint16_t v12v;
    uint16_t v1v2vdn;
    uint16_t v1v2vdd;
    uint16_t v2v5;
    uint16_t v2v5vdp;
    uint16_t v2v5vde;
    uint16_t v3v3;
} x10_sitara2_voltages_t;

uint16_t CONV_SITARA_VOLTAGE(uint16_t v, uint8_t mult)
{
  return (uint16_t)(((float)(v*25000*mult))/2560.0f);
}

int x10i2c_conv_sitara2_voltages(x10i2c_t *h, uint8_t *v, x10_sitara2_voltages_t *vl)
{
  vl->pw0 = CONV_SITARA_VOLTAGE(v[5],6);
  vl->pw1 = CONV_SITARA_VOLTAGE(v[6],6);
  vl->v12v = CONV_SITARA_VOLTAGE(v[7],6);
  vl->v1v2vdn = CONV_SITARA_VOLTAGE(v[10],1);
  vl->v1v2vdd = CONV_SITARA_VOLTAGE(v[11],1);
  vl->v2v5 = CONV_SITARA_VOLTAGE(v[12],2);
  vl->v2v5vdp = CONV_SITARA_VOLTAGE(v[13],2);
  vl->v2v5vde = CONV_SITARA_VOLTAGE(v[14],2);
  vl->v3v3 = CONV_SITARA_VOLTAGE(v[15],2);
  return 0;
}

int _x10_get_88e1111_reg(x10i2c_t *h, uint8_t i2caddr, uint8_t reg, uint16_t *val)
{
   uint8_t v=0;
   (*val)=0;
   if (x10i2c_get_reg(h, i2caddr, reg, &v))
     return -1;
   (*val)|=(v<<8);

   if (x10i2c_get_reg(h, i2caddr, reg, &v))
     return -1;
   (*val)|=v;

   return 0;
}

int _x10_set_88e1111_reg(x10i2c_t *h, uint8_t i2caddr, uint8_t reg, uint16_t val)
{
   if (x10i2c_set_reg(h, i2caddr, reg, (val>>8)))
     return -1;

   if (x10i2c_set_reg(h, i2caddr, reg, val&0xFF))
     return -1;

   return 0;
}

int x10_get_88e1111_reg(x10i2c_t *h, uint8_t i2caddr, uint8_t page, uint8_t reg, uint16_t *val)
{
  X10I2CDBG("p:%u r:%u\n", (uint16_t)page, (uint16_t)reg);
  if (reg<22)
    if (_x10_set_88e1111_reg(h, i2caddr, 22, page))
      return -1;

  return _x10_get_88e1111_reg(h, i2caddr, reg, val);
}

int x10_set_88e1111_reg(x10i2c_t *h, uint8_t i2caddr, uint8_t page, uint8_t reg, uint16_t val)
{
  X10I2CDBG("p:%u r:%u v:%04X\n", (uint16_t)page, (uint16_t)reg, (uint16_t)val);
  if (reg<22)
    if (_x10_set_88e1111_reg(h, i2caddr, 22, page))
      return -1;

  return _x10_set_88e1111_reg(h, i2caddr, reg, val);
}

int x10_get_sfp_link_status(x10i2c_t *h, uint16_t lp, uint8_t *sts)
{
  uint16_t val;
  X10I2CDBG("prt: %02d \n", (uint16_t)lp);

  if (openi2c(h, X10I2C_MAIN_BUS))
    goto error;

  if (set_pca9544_multiplexer(h, PCA9544ADDR(lp), PCA9544VALUE_SFP(lp)))
    goto error;

  if (x10_get_88e1111_reg(h, 0x56, 0, 17, &val))
    goto error;

  X10I2CDBG("val: %04X \n", val);

  (*sts) = (val&0x400)?1:0;

  set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);

  closei2c(h);

  return 0;

error:

  if (h->fd >= 0) {
    set_pca9544_multiplexer(h, PCA9544ADDR(lp), 0);
    closei2c(h);
  }

  return -1;
}

int x10i2c_sitara2_print_all_sensors(x10i2c_t *h, FILE *fl)
{
  uint8_t v[16], i, sio[2];
  uint16_t f[8];
  uint16_t t[6];

  char *fmt="%-24s: %.2f %s\n";

  if (x10i2c_get_sitara2_voltages(h, v))
    return -1;

  if (x10i2c_get_sitara2_temps(h, t))
    return -1;

  if (x10i2c_get_sitara2_fans(h, f))
    return -1;

  if (x10i2c_get_sitara2_sio_values(h, sio))
    return -1;

  float cur = ((float)(v[4]*25))/(2560.0f*0.2f);
  fprintf(fl, "%-24s: %.2f %s \n", "I system", cur,"A");
  fprintf(fl, "%-24s: %.2f %s \n", "P system", cur*12.0,"W");

  if (sio[0]&SITARA_SYSIO_TCA9535_PSB0_PRESENT)
    fprintf(fl, fmt, "V in PSB0", ((float)(v[5]*25*6))/2560.0, "V");

  if (sio[0]&SITARA_SYSIO_TCA9535_PSB1_PRESENT)
    fprintf(fl, fmt, "V in PSB1", ((float)(v[6]*25*6))/2560.0, "V");

  fprintf(fl, fmt, "V 12V", ((float)(v[7]*25*6))/2560.0, "V");
  fprintf(fl, fmt, "1V2_VDN", ((float)(v[10]*25))/2560.0, "V");
  fprintf(fl, fmt, "1V2_VDD", ((float)(v[11]*25))/2560.0, "V");
  fprintf(fl, fmt, "2V5", ((float)(v[12]*25*2))/2560.0, "V");
  fprintf(fl, fmt, "2V5_VDP", ((float)(v[13]*25*2))/2560.0, "V");
  fprintf(fl, fmt, "2V5_VDE", ((float)(v[14]*25*2))/2560.0, "V");
  fprintf(fl, fmt, "3V3", ((float)(v[15]*25*2))/2560.0, "V");

  /*
   *  0 Switch Fan
   *  1 PSW0 Fan
   *  2 PSW1 Fan
   *  3 -0
   *  4 chassis fan 0
   *  5 chassis fan 1
  */

  for (i=0; i<8; ++i)
    fprintf(fl, "Fan%-21u: %u RPM\n", i, f[i]);

  fprintf(fl, "%-24s: %u.%u C\n", "Chip Temp", (t[0]>>2)&0xFF, (t[0]&0x3)*25);

  fprintf(fl,"%-24s: %u.%u C\n", "Chassis Temp1", (t[2]>>2)&0xFF, (t[2]&0x3)*25);
  fprintf(fl,"%-24s: %u.%u C\n", "Chassis Temp2", (t[3]>>2)&0xFF, (t[3]&0x3)*25);
  fprintf(fl,"%-24s: %u.%u C\n", "Chassis Temp3", (t[5]>>2)&0xFF, (t[5]&0x3)*25);

  if (sio[0]&SITARA_SYSIO_TCA9535_PSB0_PRESENT)
    if (!x10_decode_temp(h, tempr_to_code_sitara2, v[0], &i))
      fprintf(fl,"%-24s: %u C\n", "PWR0 Temp", i);

  if (sio[0]&SITARA_SYSIO_TCA9535_PSB1_PRESENT)
    if (!x10_decode_temp(h, tempr_to_code_sitara2, v[1], &i))
      fprintf(fl,"%-24s: %u C\n", "PWR1 Temp", i);


  return 0;
}

int x10i2c_print_all_sensors(x10i2c_t *h, FILE *fl)
{
  uint8_t env_temp, chip_temp;

  if (h->x10model>=X10MODEL_SITARA2)
    return x10i2c_sitara2_print_all_sensors(h, fl);

  if (x10_get_env_temp(h, &env_temp))
    return -1;

  if (x10_get_chip_temp(h, &chip_temp))
    return -1;

  fprintf(fl, "Chipset temperature: %u C\n", chip_temp);
  fprintf(fl, "Chassis temperature: %u C\n", env_temp);

  return 0;
}

int x10_set_display_light(x10i2c_t *h, uint8_t enable)
{
  if (h->x10model < X10MODEL_SITARA2)
    return -1;

  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  change_sitara_i2c_bus(h, 1);

  set_tca9535_expander(h, SITARA_SYSIO_TCA9535ADDR, 2,
    (enable)?SITARA_SYSIO_TCA9535_BACKLIGHT:0, SITARA_SYSIO_TCA9535_BACKLIGHT);

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return 0;

error:

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return -1;
}

int x10_set_fan_percent(x10i2c_t *h, uint8_t fan, uint8_t percent)
{
  if ((h->x10model < X10MODEL_SITARA2) && (fan != X10_TOBJID_SYS))
    return 0;

  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  change_sitara_i2c_bus(h, 1);

  switch (fan) {
    case X10_TOBJID_CHIP:
      set_ADT7463_fan_speed(h, SITARA_MAIN_ADT7463, 0, percent);
      break;
    case X10_TOBJID_SYS:
      if (h->x10model<X10MODEL_SITARA2)
        x10_set_coolers(h, percent);
      else {
        set_ADT7463_fan_speed(h, SITARA_CHASSIS_ADT7463, 1, percent);
        set_ADT7463_fan_speed(h, SITARA_CHASSIS_ADT7463, 0, percent);
      }
      break;
    case X10_TOBJID_PWR0:
      set_ADT7463_fan_speed(h, SITARA_MAIN_ADT7463, 2, percent);
      break;
    case X10_TOBJID_PWR1:
      set_ADT7463_fan_speed(h, SITARA_MAIN_ADT7463, 1, percent);
      break;
    default:
      goto error;
  }

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return 0;

error:

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return -1;
}


int x10_get_temp(x10i2c_t *h, uint8_t tobjid, uint8_t *value)
{
  uint16_t t[3], max;
  uint8_t v, sio[2];

  X10I2CDBG("objid %u \n",  tobjid );

  if (h->x10model <= X10MODEL_SITARA) {
    switch (tobjid) {
      case X10_TOBJID_CHIP:
        if (x10_get_ads7830_reg(h, ADS7830ADDR, 0x8C, &v))
          return -1;
        return x10_decode_temp(h, tempr_to_code, v, value);
      case X10_TOBJID_SYS:
        if (x10_get_ads7830_reg(h, ADS7830ADDR, 0xCC, &v))
          return -1;
        return x10_decode_temp(h, tempr_to_code, v, value);
      break;
      default:
        return -1;
    }
  }

  X10I2CDBG("qqa \n");

  if ((tobjid==X10_TOBJID_PWR0) || (tobjid==X10_TOBJID_PWR1))
    if (x10i2c_get_sitara2_sio_values(h, sio))
     return -1;


  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  change_sitara_i2c_bus(h, 1);


  switch (tobjid) {

    case X10_TOBJID_CHIP:
      if (x10i2c_get_ADT7463_temp(h, SITARA_MAIN_ADT7463, 0, t))
        goto error;

      (*value) = (t[0]>>2)&0xFF;
    break;

    case X10_TOBJID_SYS:

      if (x10i2c_get_ADT7463_temp(h, SITARA_MAIN_ADT7463, 2, t))
        goto error;
      if (x10i2c_get_ADT7463_temp(h, SITARA_CHASSIS_ADT7463, 0, &t[1]))
        goto error;
      if (x10i2c_get_ADT7463_temp(h, SITARA_CHASSIS_ADT7463, 2, &t[2]))
        goto error;

      (max) = 0;
      for (v=0; v<3; ++v) {
        X10I2CDBGL(1,"t[%u] %u %04x\n",  v, (t[v]>>2)&0xFF, t[v]);
        max += (t[v]>>2)&0xFF;
        /*if (max < t[v])
            (max) = t[v];*/
      }      
      (*value) = max/3; /* (max>>2)&0xFF;*/

    break;
    case X10_TOBJID_PWR0:

      if (!(sio[0]&SITARA_SYSIO_TCA9535_PSB0_PRESENT))
        goto error;
      if (x10i2c_get_ads7830_voltage(h, SITARA_ADS7830ADDR1, 0, &v))
        goto error;
      if (x10_decode_temp(h, tempr_to_code_sitara2, v, value))
        goto error;

    break;
    case X10_TOBJID_PWR1:

      if (!(sio[0]&SITARA_SYSIO_TCA9535_PSB1_PRESENT))
        goto error;
      if (x10i2c_get_ads7830_voltage(h, SITARA_ADS7830ADDR1, 1, &v))
        goto error;
      if (x10_decode_temp(h, tempr_to_code_sitara2, v, value))
        goto error;

    break;
    default:
      goto error;

  }

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return 0;

error:

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return -1;
}

int x10_is_present(x10i2c_t *h, uint8_t objid, uint8_t *present)
{
  uint8_t sio[2];

  if (h->x10model <= X10MODEL_SITARA)
    return -1;

  if ((objid!=X10_TOBJID_PWR0) && (X10_TOBJID_PWR1!=objid))
    return -1;

  if (h->x10model <= X10MODEL_SITARA)
    return -1;

  if (x10i2c_get_sitara2_sio_values(h, sio))
    goto error;

  (*present) = (sio[0]&((objid==X10_TOBJID_PWR0)?SITARA_SYSIO_TCA9535_PSB0_PRESENT:SITARA_SYSIO_TCA9535_PSB1_PRESENT));
  return 0;

error:
  return -1;
}

#include "font6x8.h"

int x10i2c_pcf8535_init(x10i2c_t *h, uint8_t i2caddr)
{
  char buf[16];

  X10I2CDBG("\n");

  if (set_slave_addr(h, i2caddr))
    return -1;

  buf[0] = i2caddr;
  buf[1] = 0x01;
  buf[2] = 0x20;
  buf[3] = 0x09;
  buf[4] = 0x0c;
  buf[5] = 0x05;
  buf[6] = 0x14;
  buf[7] = 0x01;

  buf[8] = 0x0a;
  buf[9] = 0x0a;

  buf[10] = 0x20;
  buf[11] = 0x05;
  buf[12] = 0xe4;

  if (write(h->fd, buf, 13)!=13)
    return -1;

  return 0;
}

int x10i2c_pcf8535_clear(x10i2c_t *h, uint8_t i2caddr)
{
  int i,j;
  char *buf;

  X10I2CDBG("\n");

  if (set_slave_addr(h, i2caddr))
    return -1;

  buf = malloc(192);
  if (!buf)
    return -1;

  buf[0] = i2caddr;
  buf[1] = 0x00;
  buf[2] = 0x01;
  buf[3] = 0x40;
  buf[4] = 0x80;

  if (write(h->fd, buf,5)!=5)
    goto error;

  for (j=0; j<4; j++) {

	buf[0] = 0x78;
	for (i=1; i<135; i++)
	  buf[i] = 0x00;
	if (write(h->fd, buf, 134)!=134)
	  goto error;

  }

  buf[0] = i2caddr;
  buf[1] = 0x00;
  buf[2] = 0x01;
  buf[3] = 0x40;
  buf[4] = 0x80;

  if (write(h->fd, buf, 5)!=5)
    goto error;

  free(buf);
  return 0;

error:

  free(buf);
  return -1;
}

int x10i2c_display_init(x10i2c_t *h, uint8_t i2caddr)
{
  if (x10i2c_pcf8535_init(h, i2caddr))
    return -1;
  return x10i2c_pcf8535_clear(h, i2caddr);
}

unsigned char pcf8535_inv(unsigned char ch)
{
  char i;
  char rch = 0;

  for (i=0;i<7;i++) {

	if (ch & 0x01)
	  rch = rch | 1;
	rch = rch << 1;
	ch = ch >> 1;
  }

  if (ch&0x01)
	rch = rch | 1;

  return rch;
}

int pcf8535_putchar_xy(x10i2c_t *h, uint8_t i2caddr, char x, char y, char c)
{
  char i;
  char * ptr;
  char buf[64];

  X10I2CDBGL(2, "%s: x %i y %i '%c'\n",  x, y, c);

  set_slave_addr(h, i2caddr);

  buf[0]=i2caddr;
  buf[1]=0x00;
  buf[2]=0x01;

  buf[3]=0x40+(y&0x07);	    // Y
  buf[4]=0x80;
  buf[5]=0x80+((6*x)&0x7f);	// X
  ptr = &buf[0];

  if (write(h->fd, ptr, 6)!=6)
    return -1;

  for (i=0; i<6; i++)
    buf[1+i] = pcf8535_inv(font6x8[((int)c*6)+i]);

  buf[0]=0x78;
  ptr = &buf[0];

  if (write(h->fd, ptr,7)!=7)
    return -1;

  return 0;
}

int x10_display_printf(x10i2c_t *h, char x, char y, char *fmt, ...)
{
  int l, i;
  va_list args;

  X10I2CDBG("x %i y %i \n",  x, y);

  if ((h->x10model < X10MODEL_SITARA2))
    return -1;

  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  change_sitara_i2c_bus(h, 1);

  va_start(args, fmt);
  vsnprintf(h->s, sizeof(h->s), fmt, args);
  va_end(args);

  l = strlen(h->s);
  for (i=0 ; i<l; ++i)
    if ((x+i)<=20)
      pcf8535_putchar_xy(h, SITARA_LCD_ADDR, x+i, y, h->s[i]);

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return 0;

error:

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return -1;
}

int x10_display_clear(x10i2c_t *h)
{
  X10I2CDBG("%s:\n", __FUNCTION__);

  if ((h->x10model < X10MODEL_SITARA2))
    return -1;

  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  change_sitara_i2c_bus(h, 1);

  x10i2c_pcf8535_clear(h, SITARA_LCD_ADDR);

  restore_sitara_i2c_bus(h);
  closei2c(h);
  return 0;

error:

  restore_sitara_i2c_bus(h);
  closei2c(h);

  return -1;
}

int x10_display_clear_line(x10i2c_t *h, char y)
{
  return x10_display_printf(h, 0, y, "                     ");
}

int x10_hwtest_display(x10i2c_t *h, FILE *fl)
{
  uint8_t x,y;
  uint16_t timer=2000;
  X10I2CDBG("%s:\n", __FUNCTION__);     
  x10_display_clear(h);
  x10_set_display_light(h, 1);

  for (y=0;y<4;++y) {
    for (x=0;x<11;++x) {
      if (x<=9) x10_display_printf(h, 9-x,y, "\x80");
      x10_display_printf(h, 10+x,y, "\x80");
      usleep(100000);
    }
  }
  usleep(1000*timer);
  x10_set_display_light(h, 0);
  usleep(1000*timer);
  x10_set_display_light(h, 1);
  x10_display_clear(h);
  return 0;
}

int x10_hwtest_leds(x10i2c_t *h, FILE *fl)
{
  X10I2CDBG("%s:\n", __FUNCTION__);
  uint16_t timer=1000;
  x10_test_ports_leds(h, timer);
  usleep(1000*timer);
  x10_powerled_pin(h, 1);
  usleep(1000*timer);
  x10_powerled_pin(h, 0);
  return 0;
}

int x10_hwtest_leds_link(x10i2c_t *h, FILE *fl)
{
  uint8_t p;
  for (p = 1; p <= 24; ++p)
    x10i2c_set_port_leds(h, p, X10_TCA9535PIN_LINK_LED);
  return 0;
}

int x10_hwtest_leds_act(x10i2c_t *h, FILE *fl)
{
  uint8_t p;
  for (p = 1; p <= 24; ++p)
    x10i2c_set_port_leds(h, p, (h->x10model>2)?X10_PORTLED_ACT_NOBLINK:X10_TCA9535PIN_ACT_LED);
  return 0;
}

int x10_hwtest_leds_alm(x10i2c_t *h, FILE *fl)
{
  uint8_t p;
  for (p = 1; p <= 24; ++p)
    x10i2c_set_port_leds(h, p, X10_TCA9535PIN_ALM_LED);
  return 0;
}

int check_fan(FILE *f, char *s,  uint16_t v, uint16_t sv)
{
   /*uint16_t r=300;
   uint16_t l = (sv<r)?sv:sv-r;*/

   if (sv) {
     if (v < 100) {
       if (f)
         fprintf(f, "ERR: Fan '%s' speed %u < 100 RPM \n", s, v);
       return -1;
     }
   } else {
       if (v > 100) {
         if (f)
           fprintf(f, "ERR: Fan '%s' speed %u > 100 RPM \n", s, v);
         return -1;
       }
   }
/*
   if ((v < l) || (v > (sv+r))) {
     if (f)
       fprintf(f, "ERR: Fan '%s' speed out of range %u (%u-%u) RPM\n", s, v, l, sv+r);
     return -1;
   }
*/

   return 0;
}

char *fanstr[7]={"chip", "pw0", "pw1", "sys0", "sys1", "sys2", "sys3"};

#define HWTEST_FANS_TIMER 2000

int x10_hwtest_fans(x10i2c_t *h, FILE *fl)
{
  uint16_t f[8];
  uint8_t i;
  int r;

  /* Check speed 100%*/

  for (i=0;i<4;++i)
    x10_set_fan_percent(h, i, 100);

  usleep(HWTEST_FANS_TIMER*1000);

  if (x10i2c_get_sitara2_fans(h, f))
    return -1;

  r=0;
  r|=check_fan(fl, fanstr[0], f[0], 5200);
  for (i=1;i<6;++i)
    r|=check_fan(fl, fanstr[i], f[i+1], 9000);

  /* Check speed 0%*/

  for (i=0;i<4;++i)
    if (x10_set_fan_percent(h, i, 0))
      return -1;

  usleep(HWTEST_FANS_TIMER*1000);

  r=0;
  r|=check_fan(fl, fanstr[0], f[0], 0);
  for (i=1;i<6;++i)
    r|=check_fan(fl, fanstr[i], f[i+1], 0);

  for (i=0;i<4;++i)
    if (x10_set_fan_percent(h, i, 50))
      return -1;

  return r;
}

int check_temp(FILE *f, char *s, uint16_t val)
{
  float fval=val/=1000;
  if ((val<10.0) || (val>80.0)) {
    if (f)
      fprintf(f, "ERR: Temperature %s out of range %.2f (%.2f-%.2f) mC\n", s, fval, 10.0, 80.0);
    return -1;
  }
  return 0;
}

int x10_hwtest_temps(x10i2c_t *h, FILE *fl)
{
  uint8_t v[16], sio[2];
  uint16_t t[6];
  int r=0;
  x10_sitara2_temps_t st;

  X10I2CDBG("%s:\n", __FUNCTION__);

  if (x10i2c_get_sitara2_sio_values(h, sio))
    return -1;

  if (x10i2c_get_sitara2_voltages(h, v))
    return -1;

  if (x10i2c_get_sitara2_temps(h, t))
    return -1;

  if (x10_conv_sitara2_temps(h, t, v, sio, &st))
    return -1;

  X10I2CDBG("222\n", __FUNCTION__);

  r|=check_temp(fl, "chip", st.chip);
  r|=check_temp(fl, "sys1", st.sys[0]);
  r|=check_temp(fl, "sys2", st.sys[1]);
  r|=check_temp(fl, "sys3", st.sys[2]);

  if (sio[0]&SITARA_SYSIO_TCA9535_PSB0_PRESENT)
    r|=check_temp(fl, "pw0", st.pw[0]);
  if (sio[0]&SITARA_SYSIO_TCA9535_PSB1_PRESENT)
    r|=check_temp(fl, "pw1", st.pw[1]);

  return r;
}

int check_voltage(FILE *f, char *s, uint16_t val, uint16_t v)
{
  uint16_t r;
  r = (uint16_t)((float)v*.05f);
  if ( ((val)<(v-r)) || ((val)>(v+r)) ) {
    if (f)
      fprintf(f, "ERR: Voltage %s out of range %u (%u-%u) mV\n", s, val, v-r, v+r);
    return -1;
  }
  return 0;
}

int x10_hwtest_voltages(x10i2c_t *h, FILE *fl)
{
  uint8_t v[16], sio[2];
  x10_sitara2_voltages_t vl;

  X10I2CDBG("%s:\n", __FUNCTION__);

  if (x10i2c_get_sitara2_sio_values(h, sio))
    return -1;

  if (x10i2c_get_sitara2_voltages(h, v))
    return -1;

  x10i2c_conv_sitara2_voltages(h, v, &vl);

  int r=0;
  if (sio[0] & SITARA_SYSIO_TCA9535_PSB0_PRESENT)
    r |= check_voltage(fl, "pw0", vl.pw0, 12000);

  if (sio[0] & SITARA_SYSIO_TCA9535_PSB0_PRESENT)
    r |= check_voltage(fl, "pw1", vl.pw1, 12000);

  r |= check_voltage(fl, "v12", vl.v12v, 12000);
  r |= check_voltage(fl, "v1.2vdn", vl.v1v2vdn, 1200);
  r |= check_voltage(fl, "v1.2vdd", vl.v1v2vdd, 1200);
  r |= check_voltage(fl, "v2.5", vl.v2v5, 2500);
  r |= check_voltage(fl, "v2.5vdp", vl.v2v5vdp, 2500);
  r |= check_voltage(fl, "v2.5vde", vl.v2v5vde, 2500);
  r |= check_voltage(fl, "v3.3", vl.v3v3, 3300);

  return r;
}

int x10_check_i2cbus(x10i2c_t *h, uint8_t i2cbus, FILE *fl)
{
  if (openi2c(h, i2cbus)) {
    if (fl)
      fprintf(fl, "ERR: I2C bus %u not found\n", i2cbus);
    return -1;
  }
  closei2c(h);
  return 0;
}

int x10_hwtest_i2cbuses(x10i2c_t *h, FILE *fl)
{
  int r=0;
  r |= x10_check_i2cbus(h, X10I2C_MAIN_BUS, fl);
  r |= x10_check_i2cbus(h, X10I2C_SITARA_BUS, fl);
  return r;
}

typedef int(*hw_test_i2c_dev_fun_t)(x10i2c_t *h, uint8_t addr, FILE *fl);

typedef struct x10_i2cdev_s {
   char *name;
   uint8_t addr;
   hw_test_i2c_dev_fun_t checkfun;
} x10_i2cdev_t;


int x10_check_i2cdev(x10i2c_t *h, x10_i2cdev_t *d, FILE *fl)
{
  union i2c_smbus_data data;

  if (fl)
    fprintf(fl, "Checking: i2c: %u device: %02X '%s'...\n", h->current_bus, d->addr, d->name);

  if (d->checkfun) {
    int r=d->checkfun(h, d->addr, fl);
    if ((r) && (fl))
      fprintf(fl, "ERR: i2c %u device %02X '%s' check failed!\n", h->current_bus, d->addr, d->name);
    return r;
  }

  if (set_slave_addr(h, d->addr))
    return -1;

  if (i2c_smbus_access(h->fd,I2C_SMBUS_READ,0,I2C_SMBUS_BYTE,&data)<0) {
    X10I2CDBG("%s\n",  strerror(errno));
    if (fl)
      fprintf(fl, "ERR: i2c %u device %02X '%s' not present!\n", h->current_bus, d->addr, d->name);
    return -1;
  }

  return 0;
}

int check_pca9544(x10i2c_t *h, uint8_t addr, FILE *fl)
{
  uint8_t val;
  return get_pca9544_multiplexer(h, addr, &val);
}

int check_ADT7463(x10i2c_t *h, uint8_t addr, FILE *fl)
{
  uint8_t val;
  if (x10i2c_get_reg(h, addr, 0x3D, &val))
    return -1;
  if (val!=0x27)
    return -1;
  return 0;
}

int check_tca9535(x10i2c_t *h, uint8_t addr, FILE *fl)
{
  uint8_t val;
  if (get_tca9535_expander(h, addr, 0, &val))
    return -1;
  return 0;
}

int check_tca9535_port(x10i2c_t *h, uint8_t addr, FILE *fl)
{
  uint8_t val;

  /* preinit expander */
  x10i2c_init_port_tca9535_expander(h, addr);

  if (get_tca9535_expander(h, addr, 0, &val))
    return -1;

  if (val!=0x80) {
    fprintf(fl, "WARN: tca9535 %u reg0 %02X != 80\n", addr, val);
    return -1;
  }

  if (get_tca9535_expander(h, addr, 1, &val))
    return -1;

  if (val!=0xC0) {
    fprintf(fl, "WARN: tca9535 %u reg1 %02X != C0\n", addr, val);
    return -1;
  }

  return 0;
}

int check_ADS7830(x10i2c_t *h, uint8_t addr, FILE *fl)
{
  uint8_t val;
  if (x10i2c_get_ads7830_voltage(h, addr, 0, &val))
    return -1;
  return 0;
}

int check_pcf8535(x10i2c_t *h, uint8_t addr, FILE *fl)
{
  return x10i2c_display_init(h, addr);
}

int check_fujitsu(x10i2c_t *h, uint8_t addr, FILE *fl)
{
  uint32_t reg;
  if (x10i2c_fujitsu_get_mainreg(h, addr, 0, &reg))
    return -1;

  if (reg!=0x08AA0511)
    return -1;
  return 0;
}

int check_pca9552(x10i2c_t *h, uint8_t addr, FILE *fl)
{
  uint8_t val;
  return get_pca9552_reg(h, 0, &val);
}

int check_pll(x10i2c_t *h, uint8_t addr, FILE *fl)
{
  uint8_t val;
  return x10i2c_get_reg(h, addr, 0, &val);
}

int check_eeprom(x10i2c_t *h, uint8_t addr, FILE *fl)
{
  char buf[16];
  return read_eeprom(h, addr, 0, 16, buf);
}

char sPCA9544[]="PCA9544";
char sTCA9535[]="TCA9535";

x10_i2cdev_t sitara2_bus0_devs[]={
    {sPCA9544, 0x70, check_pca9544},
    {"EEPROM", 0x51, check_eeprom}
};

x10_i2cdev_t sitara2_bus1_devs[]={
    {sPCA9544, 0x70, check_pca9544},
    {"FUJITSU", 0x51, check_fujitsu},
    {"ADS7830", 0x48, check_ADS7830},
    {"ADS7830", 0x49, check_ADS7830},
    {sTCA9535, 0x27, check_tca9535},
    {"ADT7463", 0x2D, check_ADT7463},
    {"ADT7463", 0x2E, check_ADT7463},
    /*{"", 0x0C, 0},*/
    {"LCD", 0x3C, check_pcf8535}
};

x10_i2cdev_t sitara2_main_bus_devs[]={
    {"PLL", 0x70, check_pll},
    {sPCA9544, 0x71, check_pca9544},
    {sPCA9544, 0x72, check_pca9544},
    {sPCA9544, 0x73, check_pca9544},
    {sPCA9544, 0x74, check_pca9544},
    {sPCA9544, 0x75, check_pca9544},
    {sPCA9544, 0x76, check_pca9544},
    {"VARSOM", 0x12, 0},
    /*{"SA", 0x48, 0},*/
    {"VARSOM", 0x1B, 0},
    {"VARSOM", 0x2D, 0}
};

int x10_check_i2cdevs(x10i2c_t *h, x10_i2cdev_t *table, uint8_t count, FILE *fl)
{
  int r=0,i;
  for (i=0; i<count; ++i)
    if (x10_check_i2cdev(h, &table[i], fl)) {
      r=1;
  }
  return r;
}

int x10_hwtest_i2cdevices(x10i2c_t *h, FILE *fl)
{
  uint8_t i;
  int r=0;

  if (openi2c(h, X10I2C_SITARA_BUS))
    goto error;

  if (x10_check_i2cdev(h, &sitara2_bus0_devs[0], fl))
    goto error;

  if (fl)
      fprintf(fl, "Changed bus to 1/1\n");

  change_sitara_i2c_bus(h, 0);

  if (x10_check_i2cdevs(h, sitara2_bus0_devs,
                        sizeof(sitara2_bus0_devs)/sizeof(x10_i2cdev_t), fl))
    goto error;

  restore_sitara_i2c_bus(h);

  if (fl)
      fprintf(fl, "Changed bus to 1/2 (SERVICE) \n");

  change_sitara_i2c_bus(h, 1);

  if (x10_check_i2cdevs(h, sitara2_bus1_devs,
                        sizeof(sitara2_bus1_devs)/sizeof(x10_i2cdev_t), fl))
    goto error;

  restore_sitara_i2c_bus(h);

  closei2c(h);

  if (fl)
      fprintf(fl, "Changed bus to 2 (MAIN)\n");

  if (openi2c(h, X10I2C_MAIN_BUS))
     goto error;

  if (x10_check_i2cdevs(h, sitara2_main_bus_devs,
                        sizeof(sitara2_main_bus_devs)/sizeof(x10_i2cdev_t), fl))
     goto error;

   for (i = 0x71; i <= 0x76; i++) {

      set_pca9544_multiplexer(h, i, 4);

      if (fl)
        fprintf(fl, "Checking devices for ports %u-%u\n",
                ((i-0x71)<<2)+1, ((i-0x71)<<2)+4);

      if (check_tca9535_port(h, TCA9535ADDR, fl)) {
        r|=1;
        if (fl)
          fprintf(fl, "i2c %u device %02X '%s' check failed (port: %u)!\n",
                  h->current_bus, TCA9535ADDR, "TCA9535", ((i-0x71)<<2)+1);
      }      

      set_pca9544_multiplexer(h, i, 6);

      if (check_pca9552(h, PCA9552_ADDR, fl)) {
        r|=1;
        if (fl)
          fprintf(fl, "i2c %u device %02X '%s' check failed (port: %u)!\n",
                  h->current_bus, TCA9535ADDR, "PCA9552", ((i-0x71)<<2)+1);
      }

      if (check_tca9535_port(h, TCA9535ADDR, fl)) {
        r|=1;
        if (fl)
          fprintf(fl, "i2c %u device %02X '%s' check failed (port: %u)!\n",
                  h->current_bus, TCA9535ADDR, "TCA9535", ((i-0x71)<<2)+3);
      }

      set_pca9544_multiplexer(h, i, 0);
  }

  if (r)
    goto error;

  closei2c(h);

  return 0;

error:

  closei2c(h);

  return -1;
}

int x10_hwtest_init(x10i2c_t *h, FILE *fl)
{
  return x10i2c_init_hw(h);
}

int x10_hwtest_eeprom(x10i2c_t *h, FILE *fl)
{
  uint8_t buf[32];
  if (x10_read_eeprom(h, 0xFFE0, 32, buf)) {
    if (fl)
      fprintf(fl, "EEPROM read failed!\n");
    return -1;
  }
  if (x10_write_eeprom(h, 0xFFE0, 32, buf)) {
    if (fl)
      fprintf(fl, "EEPROM write failed!\n");
    return -1;
  }
  return 0;
}

char x10_mac[]={0x00, 0x21, 0xCE, 0x10, 0x10, 0x10};


int open_fujtsu_socket(x10i2c_t *h, char *ifname, uint8_t *mac, struct sockaddr_ll *sa)
{
  int s=-1;
  struct ifreq ifr;
  int ifindex = 0;

  memset(&ifr, 0, sizeof(ifr));

  s = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
  if (s == -1) {
    X10I2CERR("socket():");
    goto error;
  }
  /*retrieve ethernet interface index*/

  strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
  if (ioctl(s, SIOCGIFINDEX, &ifr) == -1) {
    X10I2CERR("SIOCGIFINDEX");
    goto error;
  }

  ifindex = ifr.ifr_ifindex;

  /*retrieve corresponding MAC*/
  if (ioctl(s, SIOCGIFHWADDR, &ifr) == -1) {
    X10I2CERR("SIOCGIFHWADDR");
    goto error;
  }

  memcpy(mac, ifr.ifr_hwaddr.sa_data, 6);

  /* Bring up inteface if needed */
  if (ioctl(s, SIOCGIFFLAGS, &ifr)== -1)
    X10I2CERR("SIOCGIFFLAGS");
  else {
    if (!(ifr.ifr_ifru.ifru_flags&IFF_UP)) {
      ifr.ifr_ifru.ifru_flags |= IFF_UP;
      if (ioctl(s, SIOCSIFFLAGS, &ifr) == -1)
        X10I2CERR("SIOCSIFFLAGS");
    }
  }

  /*prepare sockaddr_ll*/
  memset(sa, 0, sizeof(struct sockaddr_ll));
  sa->sll_family   = PF_PACKET;
  sa->sll_protocol = htons(0x88B5);
  sa->sll_ifindex  = ifindex;
  sa->sll_halen    = 6;

  if ( bind(s, (struct sockaddr *)sa, sizeof(struct sockaddr_ll)) == -1) {
    X10I2CERR("bind():");
    goto error;
  }

  return s;

error:
  if (s>=0) close(s);
  return -1;
}


int x10_hwtest_fujtsu(x10i2c_t *h, FILE *fl)
{
  uint8_t mac[6];
  uint8_t buf[128];
  uint16_t *u16, cnt;
  struct sockaddr_ll sa;
  int s,r;

  s = open_fujtsu_socket(h, "eth0", mac, &sa);

  X10I2CDBG("mac: %02X:%02X:%02X:%02X:%02X:%02X\n",
            mac[0],mac[1],mac[2], mac[3],mac[4],mac[5]);

  if (s<0)
    goto error;

  /* ethernet header */
  memset(buf, 0, 128);
  memcpy(buf, x10_mac, 6);
  memcpy(buf+6, mac, 6);

  u16 = (uint16_t *)(buf+12);
  (*u16) = htons(0x88B5);

  /* fujitsu header */
  u16 = (uint16_t *)(buf+14);
  (*u16) = htons(2);
  u16+=2;
  (*u16) = htons(0); /*id */
  u16++;
  (*u16) = htons(0);

  sendto(s, buf, 60, 0,
          (struct sockaddr*)&sa, sizeof(sa));

  cnt=4*5;
  while (cnt--) {

    usleep(250000);
    r = recvfrom(s, buf, 128, MSG_DONTWAIT, NULL, NULL);

    if ( (r==-1) & (errno!=EAGAIN) ) {
      X10I2CERR("recvfrom():");
      goto error;
    }

    X10I2CDBGL(3, "r:%i\n", r);

    if (r>=16) {
      u16 = (uint16_t*)(buf+12);
      X10I2CDBGL(3, "proto:%04X\n", ntohs(*u16));
      if ( ntohs(*u16) == 0x88B5 ) {
        u16++;
        X10I2CDBGL(3, "id:%04X\n", ntohs(*u16));
        if (ntohs(*u16) == (2|1)) {
          close(s);
          return 0;
        }
      }
    }
  }

error:

  if (s>=0)
    close(s);

  return -1;
}

typedef int(*hw_test_fun_t)(x10i2c_t *h, FILE *fl);

typedef struct hw_test_table_item_s {
  x10_hwtest_id_t id;
  char *name;
  hw_test_fun_t testfun;
  uint8_t flags;
} hw_test_table_item_t;

hw_test_table_item_t sitara2_hwtests[]= {
    {X10_HW_TEST_DISPLAY, "LCD Display", x10_hwtest_display, 1},
    {X10_HW_TEST_LEDS, "Front panel LEDS", x10_hwtest_leds, 1},
    {X10_HW_TEST_LINK_LEDS, "Front panel LINK LEDS", x10_hwtest_leds_link, 1},
    {X10_HW_TEST_ACT_LEDS, "Front panel ACT LEDS", x10_hwtest_leds_act, 1},
    {X10_HW_TEST_ALM_LEDS, "Front panel ALM LEDS", x10_hwtest_leds_alm, 1},    
    {X10_HW_TEST_I2CBUS, "I2C bus", x10_hwtest_i2cbuses, 0},
    {X10_HW_TEST_I2CDEVICES, "I2C devices", x10_hwtest_i2cdevices, 0},
    {X10_HW_TEST_INIT, "Init", x10_hwtest_init, 0},
    {X10_HW_TEST_TEMPS, "Temperatures test",  x10_hwtest_temps, 0},
    {X10_HW_TEST_VOLTAGES, "Voltages test", x10_hwtest_voltages, 0},
    {X10_HW_TEST_FANS, "Fans test", x10_hwtest_fans, 1},
    {X10_HW_TEST_EEPROM, "EEPROM R/W", x10_hwtest_eeprom, 0},
    {X10_HW_TEST_FUJITSU, "FUJITSU I/O", x10_hwtest_fujtsu, 1}
};

int x10_hwtest_make(x10i2c_t *h, hw_test_table_item_t *test, FILE *fl)
{
  int r=0;

  if (h->x10model<3)
    return -1;

  if (!test->testfun)
    return -1;

  r = test->testfun(h, fl);

  if ((fl) && (!(test->flags&1)))
    fprintf(fl, "Test '%s' %s\n", test->name, (!r)?"OK":"Failed");

  return r;
}

int x10_hwtest(x10i2c_t *h, x10_hwtest_id_t id, FILE *fl)
{
  unsigned int i;
  hw_test_table_item_t *tests;
  tests = sitara2_hwtests;

  for (i=0; i<sizeof(sitara2_hwtests)/sizeof(hw_test_table_item_t); ++i)
    if (id==tests[i].id)
      break;

  if (i>=sizeof(sitara2_hwtests)/sizeof(hw_test_table_item_t))
    return -1;

  return x10_hwtest_make(h, &tests[i], fl);
}

int x10_hwtest_auto(x10i2c_t *h, FILE *fl, uint8_t stop)
{
  X10I2CDBG("%s:\n", __FUNCTION__);
  hw_test_table_item_t *tests;
  int r = 0;
  unsigned int i;
  tests = sitara2_hwtests;

  for (i=0; i<sizeof(sitara2_hwtests)/sizeof(hw_test_table_item_t); ++i)
      if (!(tests[i].flags&1) && (tests[i].testfun)) {
      r = x10_hwtest_make(h, &tests[i], fl);
      if ((stop) && (r))
        return r;
    }
  return r;
}

